"""This module defines ``Portfolo`` and ``Portfolios`` classes that are used for portfolio analysis.
"""

from pyanomaly.globals import *
from pyanomaly.tcost import TransactionCost

class Portfolio:
    """Portfolio class.

    This class makes a portfolio from positions and evaluate it.

    Position information is saved in `position` attribute and portfolio information is saved in `value` attribute.
    If positive weights of the positions on a date don't add up to 1, (1 - sum(positive weights)) will be assumed to be
    invested in a risk free asset, and its information is saved in `fposition` attribute.
    The transaction cost is assumed to be 0 for risk-free assets.
    Once the portfolio is evaluated by calling ``Portfolio.eval()``, `performance` attribute is generated.

    Args:
        name: Portfolio name.
        position: Position DataFrame. It should have index = 'date' and columns = 'id' (security ID) 'ret' (return),
            and 'wgt' (portfolio weight). If it has other columns such as price, they will be kept in the `position`
            attribute. If it has 'rf' (risk-free rate) column, its values are used as risk-free rates.
        rf: Risk-free rate DataFrame with index = 'date' and columns = 'rf'. The `rf` has priority over 'rf' column in
            `position`. If `rf` = None and `position` does not have 'rf' column, the risk-free is assumed to be 0.
        pfval0: Initial portfolio value. Default to 1.
        costfcn: `TransactionCost` class, a transaction cost function, or value. If a const transaction cost of 20
            basis points is assumed, this can be set to 0.002.
        keep_position: If False, position information (`position`) is deleted after the portfolio is created.

    Attributes:
        name: Portfolio name.
        position: Position DataFrame. Its index is 'date' and has the following columns:

            - 'id': security ID provided by the user.
            - 'ret': Return between t-1 and t.
            - 'exret': Excess return over risk-free rate between t-1 and t.
            - 'wgt`: Weight at the beginning of t.
            - 'val1': Value at t.
            - 'val': Value at the beginning of t.
            - 'val0': value at t-1.
            - 'cost': Transaction cost incurred at the beginning of t.
            - Other columns: Any columns that are in the input position data.

            In the description above, 'at the beginning of t` means at t-1 after rebalancing.

        value: Portfolio value DataFrame. Its index is 'date' and has the following columns:

            - 'ret': Return between t-1 and t. This can be either net return or gross return depending on
              the evaluation method: see ``Portfolio.eval()``.
            - 'exret': Excess return over risk-free rate between t-1 and t. This can be either net excess return or
              gross excess return depending on the evaluation method.
            - 'val1': Value at t.
            - 'val': Value at the beginning of t.
            - 'cost': Transaction cost incurred at the beginning of t.
            - 'tover': Turnover incurred at the beginning of t. `tover` = sum(pos.val - pos.val0) / sum(pf.val),
              where pos.val is the position values and pf.val is the portfolio value.
            - 'netret': Return between t-1 and t, net of transaction cost.
            - 'grossret': Gross return between t-1 and t.
            - 'netexret': Excess return between t-1 and t, net of transaction cost.
            - 'grossexret': Excess gross return between t-1 and t.
            - 'lposition': Number of long positions.
            - 'sposition': Number of short positions.

            Columns added once the portfolio is evaluated by calling ``Portfolio.eval()``:

            - 'cumret': Cumulative return since the first date.
            - 'drawdown': Drawdown.
            - 'drawdur': Duration of daawdown in the frequency of data, e.g., if rebalanced monthly, 3 means 3 months.
            - 'drawstart': Beginning date of drawdown.
            - 'succdown': Successive down; down without any up during the period.
            - 'succdur': Duration of successive down.
            - 'succstart' Beginning date of successive down.

        fposition: Risk-free asset position DataFrame. Its index is 'date' and has the following columns:

            - 'ret': Return between t-1 and t.
            - 'wgt`: Weight at the beginning of t.
            - 'val1': Value at t.
            - 'val': Value at the beginning of t.

        performance: Portfolio performance DataFrame. Its column is equal to the portfolio name and has the following
            indexes:

            - 'mean': Mean excess return over the evaluation period.
            - 'std': Standard deviation of the excess returns over the evaluation period.
            - 'sharpe': Sharpe ratio.
            - 'cum': : Cumulative return over the evaluation period.
            - 'mdd': Maximum drawdown.
            - 'mdd start': Maximum drawdown start date.
            - 'mdd end': Maximum drawdown end date.
            - 'msd': Maximum successive down.
            - 'msd start': Maximum successive down start date.
            - 'msd end': Maximum successive down end date.
            - 'turnover': Average turnover.
            - 'lposotion': Average number of long positions.
            - 'sposition': Average number of short positions.

        costfcn: `TransactionCost` object, a transaction cost function, or value. For example, if a const transaction
            cost of 20 basis points is assumed, `costfcn` can be simply set to 0.002.

            If `costfcn` is a `TransactionCost` object, ``TransactionCost.get_cost(position)`` is called to get
            transaction costs. `TransactionCost` allows transaction costs varying across time and securities.
            See ``pyanomaly.tcost`` module for more details.

            When `costfcn` is defined as a function, the transaction cost function should have arguments
            val (value after rebalancing) and val0 (value before rebalancing). For example, if the transaction cost
            to buy (sell) is 20 (30) bps, a function can be defined as follows:

            .. code-block::

                def cost_fcn(val, val0):
                    return 0.002 if val > val0 else 0.003

    NOTE:
        If a position exists at t-1 but not at t, it will be added at t with 0 weight.
        This is to compute the transaction cost.

        'val', 'val1', and 'val0' in `position` and `value` are calculated without taking transaction costs into account.
        For the value increase after transaction costs, use the cumulative return.
    """

    def __init__(self, name=None, position=None, rf=None, pfval0=1, costfcn=None, keep_position=False):
        self.name = name
        self.position = None
        self.fposition = None
        self.value = None
        self.performance = None
        self.costfcn = costfcn

        if position is not None:
            self.set_position(position, rf, pfval0, keep_position)

    @staticmethod
    def _add_sold_position(position):
        """Add sold positions.

        When a stock is excluded from the portfolio, a position with val1=0 should be added for
        correct transaction cost calculation.
        """

        dates = np.sort(position.index.get_level_values(0).unique())
        dates = pd.DataFrame(dates, columns=['date'])
        dates['date0'] = dates.date.shift(1)
        position0 = position.merge(dates, left_on='date', right_on='date0')
        position0 = position0.drop(columns=['date0']).set_index('date')
        position = position.merge(position0['id'], 'outer', on=['date', 'id'], sort=True)
        position = position.fillna(0)
        return position

    @staticmethod
    def _make_portfolio_from_weight(position, rf=None, pfval0=1):
        """Make a portfolio give position weights.
        """

        position = position.copy()
        dates = np.sort(position.index.get_level_values(0).unique())

        # Initialize risk-free position and portfolio.
        fposition = pd.DataFrame(0., columns=['ret', 'wgt', 'val1', 'val'], index=dates)
        if rf is not None:
            fposition['ret'] = rf
            position['rf'] = rf
            position['exret'] = position['ret'] - position['rf']
        elif 'rf' in position:
            fposition['ret'] = position['rf'].groupby('date').first()
            position['exret'] = position['ret'] - position['rf']
        else:
            position['exret'] = position['ret']

        portfolio = pd.DataFrame(0., columns=['ret', 'exret', 'val1', 'val'], index=dates)

        position['pwgt'] = np.where(position.wgt >= 0, position.wgt, 0)
        position['wgtret'] = position.wgt * position.ret
        position['wgtexret'] = position.wgt * position.exret

        gbd = position.groupby('date')

        fposition.wgt = 1 - gbd.pwgt.sum()

        portfolio.ret = gbd.wgtret.sum() + fposition.wgt * fposition.ret
        portfolio.exret = gbd.wgtexret.sum() + fposition.wgt * fposition.ret
        position.drop(columns=['pwgt', 'wgtret', 'wgtexret'], inplace=True)

        portfolio.val1 = pfval0 * (1 + portfolio.ret).cumprod()
        portfolio.val = portfolio.val1.shift(1)
        portfolio.val.values[0] = pfval0

        portfolio['lposition'] = position.loc[position['wgt'] > 0, 'wgt'].groupby('date').count().fillna(0)
        portfolio['lposition'].fillna(0, inplace=True)
        portfolio['sposition'] = position.loc[position['wgt'] < 0, 'wgt'].groupby('date').count().fillna(0)
        portfolio['sposition'].fillna(0, inplace=True)


        position['val'] = portfolio.val
        position['val'] *= position.wgt
        position['val1'] = position.val * (1 + position.ret)

        fposition.val = fposition.wgt * portfolio.val
        fposition.val1 = fposition.val * (1 + fposition.ret)
        fposition = fposition.fillna(0)

        position = Portfolio._add_sold_position(position)
        return portfolio, position, fposition

    @staticmethod
    def _add_transaction_cost(portfolio, position, fposition, costfcn=None):
        """Add transaction costs and turnover to position and portfolio.
        """

        position['val0'] = position.groupby('id').val1.shift(1).fillna(0)
        # Set the cost of the first date to 0
        date0 = portfolio.index.min()
        position.loc[date0, 'val0'] = position.loc[date0, 'val']
        position['dval'] = np.abs(position.val - position.val0)

        if costfcn is None:
            position['cost'] = 0
        elif callable(costfcn):
            vcostfcn = np.vectorize(costfcn)
            position['cost'] = vcostfcn(position.val, position.val0)
        elif isinstance(costfcn, TransactionCost):
            position['cost'] = costfcn.get_cost(position)
        else:  # constant transaction cost
            position['cost'] = costfcn * position.dval
        portfolio[['cost', 'tover']] = position.groupby('date')[['cost', 'dval']].sum()
        portfolio.tover = portfolio.tover / portfolio.val

        portfolio['netret'] = portfolio.ret - portfolio.cost / portfolio.val
        portfolio['grossret'] = portfolio.ret
        portfolio['netexret'] = portfolio.exret - portfolio.cost / portfolio.val
        portfolio['grossexret'] = portfolio.exret
        del position['dval']

        return portfolio, position, fposition

    def set_position(self, position, rf=None, pfval0=1, keep_position=False):
        """Set positions.

        This method sets `position` attribute from the input arguments.
        For the details of the input arguments, See the arguments of the initializer.
        """

        self.value, self.position, self.fposition = Portfolio._make_portfolio_from_weight(position, rf, pfval0)
        Portfolio._add_transaction_cost(self.value, self.position, self.fposition, self.costfcn)

        if not keep_position: self.position = None

    @staticmethod
    def from_portfolo_return(pfret, val=1):
        """Make a portfolio given portfolio returns.

        If portfolio returns are already known, this method can be used to construct a `Portfoliio` object
        and evaluate the portfolio.

        Args:
            pfret: DataFrame with index = 'date' and columns = 'ret'.

        Returns:
            `Portfolio` object.
        """

        pfval = pfret.copy()
        pfval['exret'] = pfval.ret
        pfval['val1'] = val * (1 + pfval.ret).cumprod()
        pfval['val'] = pfval.val1.shift(1)
        pfval.val.values[0] = val
        pfval['cost'] = 0
        pfval['tover'] = 0
        pfval['netret'] = pfval.ret
        pfval['grossret'] = pfval.ret
        pfval['netexret'] = pfval.exret
        pfval['grossexret'] = pfval.exret
        pfval['lposition'] = 1
        pfval['sposition'] = 0

        portfolio = Portfolio()
        portfolio.value = pfval
        return portfolio

    def add_position(self, position, rf=None):
        """Add new positions.

        This method assumes positions are added in order of time. All positions of a given date must be added together.
        The arguments have the same formats as those of the initializer.
        """

        if self.position is None:
            self.set_position(position, rf, keep_position=True)
            return

        val = self.value.val1[-1]  # val = val1 of last time
        portfolio, position, fposition = Portfolio._make_portfolio_from_weight(position, rf, val)

        date0 = self.position.index[-1]
        portfolio = pd.concat([self.value.loc[[date0]], portfolio], sort=False)
        position = pd.concat([self.position.loc[[date0]], position], sort=False)
        fposition = pd.concat([self.fposition.loc[[date0]], fposition], sort=False)
        portfolio, position, fposition = Portfolio._add_transaction_cost(portfolio, position,
                                                               fposition, self.costfcn)
        self.value = pd.concat([self.value, portfolio[portfolio.index > date0]])
        self.position = pd.concat([self.position, position[position.index > date0]])
        self.fposition = pd.concat([self.fposition, fposition[fposition.index > date0]])

    def copy(self, sdate=None, edate=None):
        """Copy this object for the given period.

        Args:
            sdate: Start date ('yyyy-mm-dd').
            edate: End date.

        Returns:
            `Portfolio` object.
        """
        pf = Portfolio()
        pf.name = self.name
        pf.value = self.value[sdate:edate].copy()
        pf.position = self.position[sdate:edate].copy()
        pf.fposition = self.fposition[sdate:edate].copy()
        pf.costfcn = self.costfcn

        return pf

    def set_return_type(self, consider_cost=True):
        """Determine which return (gross vs. net) to use for portfolio evaluation.

        If `consider_cost` is True, `value.ret` and `value.exret` are respectively set to net returns (`value.netret`)
        and net excess returns (`value.netexret`). Otherwise, they are set to gross returns (`value.grossret`) and
        gross excess returns (`value.grossexret`).

        Args:
            consider_cost: True to use returns net of transaction costs for evaluation.
        """

        if consider_cost:
            self.value.ret = self.value.netret
            self.value.exret = self.value.netexret
        else:
            self.value.ret = self.value.grossret
            self.value.exret = self.value.grossexret

    def returns(self, sdate=None, edate=None):
        """Get portfolio returns.

        Both `sdate` and `edate` are inclusive, i.e., the first return is the return over `sdate`-1 and `sdate`.

        Args:
            sdate: Start date.
            edate: End date.

        Returns:
            Return Series with index = 'date'.
        """

        return self.value[sdate:edate].ret

    @staticmethod
    def _cum_returns(ret, logscale=True, zero_start=False):
        cumret = np.log(1 + ret).cumsum()
        if zero_start: cumret = cumret - cumret.iloc[0]
        if not logscale:
            cumret = np.exp(cumret) - 1
        return cumret

    def cum_returns(self, sdate=None, edate=None, logscale=True, zero_start=False):
        """Get portfolio cumulative returns.

        Both `sdate` and `edate` are inclusive, i.e., the first return is the return over `sdate`-1 and `sdate`.

        Args:
            sdate: Start date.
            edate: End date.
            logscale: If True, return log-scale cumulative returns.
            zero_start: If True, the return at `sdate` is forced to 0. That is, it is assumed that trading starts
                at the end of `sdate`. This is useful when plotting cumulative returns as all curves will start at the
                same point, 0.

        Returns:
            Cumulative return Series with index = 'date'.
        """

        return self._cum_returns(self.value[sdate:edate].ret, logscale, zero_start)

    def mean_return(self, sdate=None, edate=None):
        """Get portfolio mean return over the period.

        Args:
            sdate: Start date.
            edate: End date.

        Returns:
            Mean return.
        """

        return self.value[sdate:edate].ret.mean()

    def std_return(self, sdate=None, edate=None):
        """Get the standard deviation of the returns over the period.

        Args:
            sdate: Start date.
            edate: End date.

        Returns:
            Standard deviation of the returns.
        """
        return self.value[sdate:edate].ret.std()

    def cum_return(self, sdate=None, edate=None, logscale=True):
        """Get the cumulative return over the period.

        Args:
            sdate: Start date.
            edate: End date.
            logscale: If True, return log-scale cumulative return.

        Returns:
            Cumulative return.
        """

        cumret = np.log(1 + self.value[sdate:edate].ret.values).sum()
        if not logscale:
            cumret = np.exp(cumret) - 1

        return cumret

    def sharpe_ratio(self, sdate=None, edate=None):
        """Get the Sharpe ratio over the period.

        If risk-free rates are not set, they are assumed to be 0.

        Args:
            sdate: Start date.
            edate: End date.

        Returns:
            Sharpe ratio.
        """

        er = self.value[sdate:edate].exret.values
        mean = er.mean()
        std = er.std()
        return mean / std if std != 0 else 0

    @staticmethod
    def _get_succdown(cumret):
        ret = np.diff(np.insert(cumret, 0, 0))
        succdown = np.zeros_like(cumret)
        duration = np.zeros_like(cumret)
        cumret0 = 0
        t0 = -1
        for t in range(len(cumret)):
            if ret[t] > 0:
                cumret0 = cumret[t]
                t0 = t
            succdown[t] = cumret0 - cumret[t]
            duration[t] = t - t0
        return succdown, duration

    @staticmethod
    def _succdown(cumret, logscale=True):
        value, duration = Portfolio._get_succdown(cumret.values)
        if not logscale: value = 1 - np.exp(-value)
        succdown = pd.DataFrame(np.transpose([value, duration]), columns=['value', 'duration'], index=cumret.index)
        idx = np.arange(len(duration)) - duration.astype(int)
        idx[idx < 0] = 0  # to prevent negative index (can happen when ret < 0 in the first month)
        succdown['start'] = cumret.index[idx]
        return succdown

    def succdown(self, sdate=None, edate=None, logscale=True):
        """Get the successive downs over the period.

        Args:
            sdate: Start date.
            edate: End date.
            logscale: If True, return log-scale values.

        Returns:
            Successive downs. DataFrame with index = 'date' and columns = ['value', 'duration', 'start'].
        """

        return self._succdown(self.cum_returns(sdate, edate), logscale)

    def max_succdown(self, sdate=None, edate=None, logscale=True):
        """Get the maximum successive down over the period.

        Args:
            sdate: Start date.
            edate: End date.
            logscale: If True, return log-scale value.

        Returns:
            Maximum successive down. Series with index = ['value', 'duration', 'start'].
        """

        succdown = self.succdown(sdate, edate, logscale)
        return succdown.loc[succdown.value.idxmax()]

    @staticmethod
    def _get_drawdown(cumret):
        drawdown = np.zeros_like(cumret)
        duration = np.zeros_like(cumret)
        cumret0 = 0
        t0 = -1
        for t in range(len(cumret)):
            if cumret[t] >= cumret0:
                cumret0 = cumret[t]
                t0 = t
            drawdown[t] = cumret0 - cumret[t]
            duration[t] = t - t0
        return drawdown, duration

    @staticmethod
    def _drawdown(cumret, logscale=True):
        value, duration = Portfolio._get_drawdown(cumret.values)
        if not logscale: value = 1 - np.exp(-value)
        drawdown = pd.DataFrame(np.transpose([value, duration]), columns=['value', 'duration'], index=cumret.index)
        idx = np.arange(len(duration)) - duration.astype(int)
        idx[idx < 0] = 0  # to prevent negative index (can happen when ret < 0 in the first month)
        drawdown['start'] = cumret.index[idx]
        return drawdown

    def drawdown(self, sdate=None, edate=None, logscale=True):
        """Get the drawdowns over the period.

        Args:
            sdate: Start date.
            edate: End date.
            logscale: If True, return log-scale values.

        Returns:
            DataFrame of drawdowns with index = 'date' and columns = ['value', 'duration', 'start'].
        """

        return self._drawdown(self.cum_returns(sdate, edate), logscale)

    def max_drawdown(self, sdate=None, edate=None, logscale=True):
        """Get the maximum drawdown over the period.

        Args:
            sdate: Start date.
            edate: End date.
            logscale: If True, return log-scale value.

        Returns:
            Maximum drawdown. Series with index = ['value', 'duration', 'start'].
        """

        drawdown = self.drawdown(sdate, edate, logscale)
        return drawdown.loc[drawdown.value.idxmax()]

    def eval(self, sdate=None, edate=None, logscale=True, annualize_factor=1, consider_cost=True, percentage=False):
        """Evaluate the portfolio over the period.

        This method evaluate the portfolio and create `performance` attribute. It also adds performance-related columns
        to `value` attribute.

        `performance`: DataFrame with columns = [`name`] and the following indexes:

            - 'mean': Mean excess return over the evaluation period.
            - 'std': Standard deviation of the excess returns over the evaluation period.
            - 'sharpe': Sharpe ratio.
            - 'cum': : Cumulative return over the evaluation period.
            - 'mdd': Maximum drawdown.
            - 'mdd start': Maximum drawdown start date.
            - 'mdd end': Maximum drawdown end date.
            - 'msd': Maximum successive down.
            - 'msd start': Maximum successive down start date.
            - 'msd end': Maximum successive down end date.
            - 'turnover': Average turnover.
            - 'lposotion': Average number of long positions.
            - 'sposition': Average number of short positions.

        Columns added to `value`:

            - 'cumret': Cumulative return since `sdate`.
            - 'drawdown': Drawdown.
            - 'drawdur': Duration of daawdown in the frequency of data, e.g., if rebalanced monthly, 3 means 3 months.
            - 'drawstart': Beginning date of drawdown.
            - 'succdown': Successive down; down without any up during the period.
            - 'succdur': Duration of successive down.
            - 'succstart' Beginning date of successive down.

        Args:
            sdate: Start date.
            edate: End date.
            logscale: If True, 'cum', 'mdd', 'msd' are in log-scale.
            annualize_factor: 'mean', 'std', 'sharpe', and 'turnover' are annualized by this factor, e.g.,
                `mean` is multiplied by `annualize_factor` and `std` by its square-root.
                If the data is monthly, the results can be annualized by setting `annualize_factor` = 12. Default to 1.
            consider_cost: If True, the results are calculated using net returns, otherwise, they are calculated
                using gross returns. If True (False), `value.ret` will be set to `value.netret` (`value.grossret`).
            percentage: If True, 'mean', 'std', 'cum', 'mdd', 'msd', and 'turnover' are multiplied by 100.
                Default to False.

        Returns:
            `performance`, `value`.
        """

        self.set_return_type(consider_cost)

        cumret = self.cum_returns(sdate, edate, logscale)
        if logscale:
            drawdown = self._drawdown(cumret, logscale=False)
            succdown = self._succdown(cumret, logscale=False)
        else:
            drawdown = self._drawdown(np.log(1 + cumret), logscale=False)
            succdown = self._succdown(np.log(1 + cumret), logscale=False)

        self.value['cumret'] = cumret
        self.value[['drawdown', 'drawdur', 'drawstart']] = drawdown
        self.value[['succdown', 'succdur', 'succstart']] = succdown

        portfolio = self.value[sdate:edate]
        r = portfolio.ret.values
        er = portfolio.exret.values
        # if self.fposition is not None:
        #     r -= self.fposition[sdate:edate].ret.values
        mean = er.mean()
        std = er.std()
        sharpe = mean / std if std != 0 else 0
        cum = cumret[-1]

        mdd = drawdown.loc[drawdown.value.idxmax()]
        msd = succdown.loc[succdown.value.idxmax()]
        dates = np.datetime_as_string(pd.to_datetime([mdd.start, mdd.name, msd.start, msd.name]), 'D')

        tover = portfolio.tover.mean()

        lposition = portfolio.lposition.mean()
        sposition = portfolio.sposition.mean()

        # Annualize
        mean *= annualize_factor
        std *= np.sqrt(annualize_factor)
        sharpe *= np.sqrt(annualize_factor)
        tover *= annualize_factor

        values = [mean, std, sharpe, cum, mdd.value, dates[0], dates[1], msd.value, dates[2], dates[3],
                  tover, lposition, sposition]
        index = ['mean', 'std', 'sharpe', 'cum', 'mdd', 'mdd start', 'mdd end', 'msd', 'msd start', 'msd end',
                 'turnover', 'lposition', 'sposition']
        self.performance = pd.DataFrame(values, index=index, columns=[self.name])
        if percentage:
            self.performance.loc[['mean', 'std', 'cum', 'mdd', 'msd', 'turnover']] *= 100
        return self.performance, self.value[sdate:edate]

    def eval_series(self, sdate=None, edate=None, logscale=True, annualize_factor=1, consider_cost=True, percentage=False):
        """Evaluate the portfolio repeatedly over the period.

        Evaluate the portfolio repeatedly for the period [`sdate`, `sdate+1`], [`sdate`, `sdate+2`], ...,
        [`sdate`, `edate`]. For the description of the arguments, see ``Portfolio.eval()``.

        Returns:
            Performance for each period. DataFrame with index values equal to the period end dates and columns
            equal to the indexes of `performance` attribute, i.e., a row with index t contains the performance up to t.
        """

        if sdate is None: sdate = self.value.index[0]
        if edate is None: edate = self.value.index[-1]
        perf_list = {}
        for date in self.value.index:
            if date <= sdate: continue
            if date > edate: break
            perf_list[date], _ = self.eval(sdate, date, logscale, annualize_factor, consider_cost, percentage)
        perf_list = pd.concat(perf_list, axis=1).transpose()
        return perf_list.droplevel(axis=0, level=1)  # drop portfolio name and keep dates only

    # def diff(self, bm):
    #     pfdiff = Portfolio()
    #     pfdiff.name = self.name + '-' + bm.name
    #     pfdiff.value = (self.value.ret - bm.value.ret).to_frame()
    #     return pfdiff


class Portfolios():
    """Class for a group of portfolios.

    This class can have several portfolios as its members and evaluate them together.
    This class facilitates portfolio comparison by evaluating them and saving the results in a single DataFrame.

    Args:
        portfolios: List or dict of `Portfolio` objects to add. If it is a dict, its keys are used as
            portfolio names. Portfolios can be added later using ``Portfolios.add()`` or ``Portfolios.set()``.

    Attributes:
        members (dict): Portfolio members. A `Portfolio` object is added to `members` with its name as the key.
            A member portfolio can be accessed by ``__getitem__()``:

            >>> pf1 = Portfolio('pf1')
            >>> portfolios = Portfolios()
            >>> portfolios.add(pf1)
            >>> pf1 = portfolios['pf1']  # This and the next line are equivalent.
            >>> pf1 = portfolios.members['pf1']

        value: Portfolios' values. This is a concatenated DataFrame of the `value` attributes of the members.
            It has the same structure as `Portfolio.value` except that its columns are two-level: the first level
            is the same as the columns of `Portfolio.value` and the second level is the member names.
        performance: Portfolios' performances. This is a concatenated DataFrame of the `performance` attributes of
            the members. It has the same structure as `Portfolio.performance` except that it has multiple columns,
            each corresponding to a member. The column names are the same as the portfolio names.
    """

    def __init__(self, portfolios=None):
        self.members = {}
        self.value = None
        self.performance = None
        if portfolios:
            self.set(portfolios)

    def __getitem__(self, alias):
        return self.members[alias]

    def add(self, portfolio, alias=None):
        """Add a portfolio.

        Args:
            portfolio (Portfolio): Portfolio to add.
            alias: Portfolio alias. If not None, this is used as the portfolio name.
        """
        key = alias if alias else portfolio.name
        self.members[key] = portfolio
        pfval = portfolio.value.copy()
        pfval.columns = pd.MultiIndex.from_product([pfval.columns, [key]])
        self.value = pd.concat([self.value, pfval], axis=1)
        if portfolio.performance is not None:
            self.performance = pd.concat([self.performance,
                                          portfolio.performance.rename(columns={portfolio.name: key})], axis=1)

    def set(self, portfolios):
        """Add portfolios.

        Args:
            portfolios: List or dict of `Portfolio` objects. If it is a dict, its keys are used as
                portfolio names.
        """

        if isinstance(portfolios, dict):
            self.members = portfolios
        elif isinstance(portfolios, list):
            self.members = {pf.name: pf for pf in portfolios}
        pfvals = []
        pfperf = []
        for portfolio in self.members.values():
            pfvals.append(portfolio.value)
            pfperf.append(portfolio.performance)
        self.value = pd.concat(pfvals, axis=1, keys=self.members.keys()).swaplevel(0, 1, axis=1)
        if not any(perf is None for perf in pfperf):
            self.performance = pd.concat(pfperf, axis=1)
            self.performance.columns = self.members.keys()

    def eval(self, sdate=None, edate=None, logscale=True, annualize_factor=1, consider_cost=True, percentage=False):
        """Evaluate the portfolios.

        For the arguments, see ``Portfolio.eval()``.

        Returns:
            `performance`, `value`.
        """

        for portfolio in self.members.values():
            portfolio.eval(sdate, edate, logscale, annualize_factor, consider_cost, percentage)
        self.set(self.members)

        return self.performance, self.value[sdate:edate]


# def compare_portfolios(portfolio, benchmark, sdate=None, edate=None, logscale=True, annualized=True,
#                        consider_cost=True):
#     diff = portfolio.diff(benchmark)
#     portfolios = Portfolios([portfolio, benchmark, diff])
#     return portfolios.eval(sdate, edate, logscale, annualized, consider_cost)





"""This module defines classes for transaction costs.
"""
import pandas as pd

from pyanomaly.globals import *


class TransactionCost:
    """Transaction cost class.

    Trnansaction cost can be set at the security level. It can also vary over time. See ``set_params()`` for details.

    Attributes:
        params: Transaction cost parameters. This can be a float number, dict, or DataFrame. See ``set_params()``.
    """

    def __init__(self, **kwargs):
        self.params = None
        if kwargs:
            self.set_params(**kwargs)

    def set_params(self, **kwargs):
        """Set transaction cost parameters.

        Parameters can be set either by this method or by ``__init__()``.

        Args:
            kwargs: `kwargs` can have the following keywords:

                - 'cost': For a constant (scalar) transaction cost.
                - 'buy_fixed`, 'buy_linear`, 'buy_quad', 'sell_fixed', 'sell_linear', 'sell_quad': For
                  a quadratic transaction cost.
                - 'params': For a transaction cost that varies across securities (and over time).

        Below are some examples.

        For a constant transaction cost of 20 basis points:

        >>> tc = TransactionCost(cost=0.002)

        Asymmetric quadratic cost function:

            `cost = fixed + linear * Amount + quad * Amount^2`

            - To buy: fixed cost = $5, linear cost = 0.002, and quadratic cost = 0.001
            - To sell: fixed cost = 0, linear cost = 0.003, and quadratic cost = 0.001

        >>> tc = TransactionCost(buy_fixed=5, buy_linear=0.002, buy_quad=0.001, sell_linear=0.003, sell_quad=0.001)

            Only non-zero parameters need to be provided.

        Transaction costs that vary across securities:

            - Security 1 (id: 0001): 0.002, security 2 (id: 0002): 0.003

        .. code-block::

            >>> params = pd.DataFrame({'cost': [0.002, 0.003]}, index=pd.Index(['0001', '0002'], name='id'))
            >>> params
                  cost
            id
            0001 0.002
            0002 0.003
            >>> tc = TransactionCost(params=params)

        Transaction costs that vary across securities and dates:

            - Security 1 (id: 0001): 0.004 on '2000-03-31', 0.003 on '2000-04-30'
            - Security 2 (id: 0002): 0.005 on '2000-03-31', 0.004 on '2000-04-30'

        .. code-block::

            >>> dates = ['2000-03-31', '2000-04-30']
            >>> ids = ['0001', '0002']
            >>> params = pd.DataFrame(index=pd.MultiIndex.from_product([dates, ids], names=('date', 'id')))
            >>> params['cost'] = [0.004, 0.005, 0.003, 0.004]
            >>> params
                             cost
            date       id
            2000-03-31 0001 0.004
                       0002 0.005
            2000-04-30 0001 0.003
                       0002 0.004
            >>> tc = TransactionCost(params=params)

        The `params` DataFrame must have index = 'id' or 'date'/'id'. It can have columns such as 'buy_fixed' instead
        of 'cost' for a more complex transaction cost structure.

        """
        if 'cost' in kwargs:
            self.params = kwargs['cost']
        elif 'params' in kwargs:
            self.params = kwargs['params']
        else:
            self.params = dict(
                buy_fixed=0,
                buy_linear=0,
                buy_quad=0,
                sell_fixed=0,
                sell_linear=0,
                sell_quad=0,
            )
            for k in self.params:
                if k in kwargs:
                    self.params[k] = kwargs[k]

    @staticmethod
    def get_cost_(val, val0, params):
        """Get a quadratic transaction cost

        Args:
            val: Value after rebalancing.
            val0: Value before rebalancing.
            params (dict): Transaction cost parameters.

        Returns:
            Transaction cost.
        """

        if val > val0:  # buy
            return params['buy_fixed'] + params['buy_linear'] * (val - val0) + params['buy_quad'] * (val - val0) ** 2
        else: # sell
            return params['sell_fixed'] + params['sell_linear'] * (val0 - val) + params['sell_quad'] * (val0 - val) ** 2

    def get_cost(self, position):
        """Get transaction costs.

        Args:
            position: Portfolio positions. ``Portfolio`` object calls this function with the argument,
                ``Portfolio.position``, to get transaction costs. The `position` argument should have index = 'date' and
                columns = [`id`, 'val', 'val0'], where 'val' is the value after rebalancing and 'val0' is the value
                before rebalancing.

        Returns:
            Transaction costs. A vector with the same length as `position`.
        """

        if self.params is None:
            return 0
        elif isinstance(self.params, float):
            return self.params * np.abs(position.val - position.val0)
        elif isinstance(self.params, dict):
            vcostfcn = np.vectorize(TransactionCost.get_cost_)
            return vcostfcn(position.val, position.val0, self.params)
        else:
            if isinstance(self.params.index, pd.MultiIndex):  # date/id
                tmp = pd.merge(position[['id', 'val', 'val0']], self.params, how='left', on=['date', 'id'])
            else:  # id
                tmp = pd.merge(position[['id', 'val', 'val0']], self.params, how='left', on='id')

            if 'cost' in tmp:
                tmp['cost'] = tmp['cost'] * np.abs(tmp['val'] - tmp['val0'])
            else:
                tmp['buy_cost'] = 0
                if 'buy_fixed' in tmp:
                    tmp['buy_cost'] += tmp['buy_fixed']
                elif 'buy_linear' in tmp:
                    tmp['buy_cost'] += tmp['buy_linear'] * (tmp['val'] - tmp['val0'])
                elif 'buy_quad' in tmp:
                    tmp['buy_cost'] += tmp['buy_quad'] * (tmp['val'] - tmp['val0']) ** 2

                tmp['sell_cost'] = 0
                if 'sell_fixed' in tmp:
                    tmp['sell_cost'] += tmp['sell_fixed']
                elif 'sell_linear' in tmp:
                    tmp['sell_cost'] += tmp['sell_linear'] * (tmp['val0'] - tmp['val'])
                elif 'sell_quad' in tmp:
                    tmp['sell_cost'] += tmp['sell_quad'] * (tmp['val0'] - tmp['val']) ** 2

                tmp['cost'] = np.where(tmp['val'] > tmp['val0'], tmp['buy_cost'], tmp['sell_cost'])

            return tmp.cost.values


class TimeVaryingCost(TransactionCost):
    """Transaction costs that vary over time and across firm sizes.

    This class implements the time-varying transaction costs used in, e.g., Brandt et al. (2009), Hand and Green (2011),
    DeMiguel et al. (2020), and Han (2021). Transaction cost parameter `k` is defined as `k` = `y` * `z`,
    where `y` decreases linearly from 4.0 in 1974.01 to 1.0 in 2002.01 and remains at 1.0 thereafter, and
    `z` = 0.006 - 0.0025 * `nme`, where `nme` is the normalized market equity that has a value between 0 and 1.

    The maximum transaction cost is 240 basis points (the smallest firm before 1974) and the minimum transaction cost is
    35 basis points (the largest firm after 2002).

    We find this assumption is too conservative since the normalized me is sensitive to the largest company.
    In 1974, the mean of the normalized me is only 0.0045 and most firms have the transaction cost of 240 basis points,
    and in 2002, the mean of the normalized me is only 0.0059 and most firms have the transaction cost of 60 basis points.

    Using a logarithm of the market equity or capping the me of the largest firms may make more sense.
    """

    def __init__(self, me=None, normalize=True):
        super().__init__()
        if me is not None:
            self.set_params(me, normalize)

    def set_params(self, me, normalize=True):
        """Set transaction cost parameters.

        Args:
            me: DataFrame or Series of market equity with index = date/id.
            normalize: If True, normalize `me` so that its values are between 0 and 1.

        NOTE:
            If the input contains only a subset of all listed stocks, normalizing the market equity can result in
            over- or underestimation of the transaction costs. For example, if `me` contains only top 80% of the stocks,
            the transaction costs will be overestimated. Use all listed stocks in the market or normalize the market equity
            outside and set `normalize = False`.
        """

        date1 = pd.Timestamp('1974-01-01')
        date2 = pd.Timestamp('2002-01-01')
        dates = me.index.get_level_values(0)

        if isinstance(me, pd.Series):
            params = me.to_frame()
        else:
            params = me.copy()
        params.index = params.index.set_names(('date', 'id'))

        if normalize:
            params['nme'] = params.groupby('date').transform(lambda x: (x - x.min()) / (x.max() - x.min()))
        else:
            params['nme'] = params.iloc[:, 0]
        params['y'] = 1 + (4 - 1) * (date2 - dates) / (date2 - date1)
        params.loc[dates < date1, 'y'] = 4.0
        params.loc[dates > date2, 'y'] = 1.0
        params['cost'] = params['y'] * (0.006 - 0.0025 * params['nme'])

        self.params = params[['cost']]




"""Fama-French Factors.

This module defines a function to generate Fama-French factors.

"""
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from pyanomaly.globals import *
from pyanomaly.datatools import *
from pyanomaly.analytics import classify, two_dim_sort, relabel_class
from pyanomaly.characteristics import FUNDA, CRSPM
from pyanomaly.wrdsdata import WRDS


def make_ff_factors():
    """Generate Fama-French 3 factors.

    This function refers to the WRDS code, but the results are slightly different as the code is written under
    the architecture of PyAnomaly. Compared to the data from the K. French website, HML has a correlation of 0.967,
    and SMB has a correlation of 0.989. Compared to the WRDS code, HML has a correlation of 0.991,
    and SMB has a correlation of 0.993.

    WRDS code:
    https://wrds-www.wharton.upenn.edu/pages/support/applications/python-replications/fama-french-factors-python/

    Major differences from WRDS code:

    1. Primary stock identification: for our method, refer to ``wrds.add_gvkey_to_crsp()``.
    2. Delist return: for our method, refer to ``wrds.merge_sf_with_seall()``.

    .. todo::
        Add mktrf, rf, and the rest of the 5 factors.

    Returns:
        factors, number of firms.

        * factors (DataFrame): index = 'date', columns: 'bh', 'bm', 'bl', 'sh', 'sm', 'sl', 'hml', 'smb'.
        * number of firms (DataFrame): index = 'date', columns: 'bh', 'bm', 'bl', 'sh', 'sm', 'sl', 'hml', 'smb'.
    """

    # CRSPM
    crspm = CRSPM()
    crspm.load_data()

    crspm.filter(('exchcd', 'in', [1, 2, 3]))
    crspm.filter(('shrcd', 'in', [10, 11]))
    crspm.update_variables()

    cm = crspm.data.reset_index()
    cm['me'] = cm.me_company  # Use company level me.
    cm['me_1'] = cm.groupby('permno').me.shift(1)  # 1-month lagged me.
    # For me_1, the following code is more rigorous as there can be missing months. But the difference seems negligible.
    # cm0 = cm[['date', 'permno', 'me']].rename(columns={'me': 'me_1'})
    # cm0['date'] = add_months(cm0.date, 1)
    # cm = cm.merge(cm0, on=['date', 'permno'])

    # Primary security only
    cm = cm[(cm.primary == True) & (cm.me > 0)]

    # FUNDA
    funda = FUNDA()
    funda.load_data()
    funda.update_variables()
    fa = funda.data.reset_index()
    fa['year'] = fa.datadate.dt.year
    fa = fa.drop_duplicates(['gvkey', 'year'], keep='last')  # Fiscal year change can cause dup. Keep the most recent.
    fa['rcount'] = fa.groupby('gvkey').cumcount()  # reset rcount (since some rows are dropped)
    fa = fa[['gvkey', 'datadate', 'be', 'rcount', 'year']]

    # Filter funda
    fa = fa[(fa.rcount >= 1) & (fa.be > 0)]

    # ME for BEME and SIZE
    # BEME: t-1 year-end me, SIZE: t year June me.
    me_jun = cm.loc[cm.date.dt.month == 6, ['date', 'gvkey', 'permno', 'me', 'exchcd']].rename(columns={'me': 'me_jun'})
    me_jun['year'] = me_jun.date.dt.year - 1
    fa = fa.merge(me_jun, on=['year', 'gvkey'])

    me_dec = cm.loc[cm.date.dt.month == 12, ['date', 'gvkey', 'permno', 'me']].rename(columns={'me': 'me_dec'})
    me_dec['year'] = me_dec.date.dt.year
    fa = fa.merge(me_dec, on=['year', 'gvkey', 'permno'])

    # me_jun = cm.loc[cm.date.dt.month == 6, ['date', 'permno', 'gvkey', 'me', 'exchcd']].rename(columns={'me': 'me_jun'})
    # me_jun['year'] = me_jun.date.dt.year - 1
    # me_dec = cm.loc[cm.date.dt.month == 12, ['date', 'permno', 'me']].rename(columns={'me': 'me_dec'})
    # me_dec['year'] = me_dec.date.dt.year
    # me = pd.merge(me_dec, me_jun, on=['year', 'permno'])
    # fa = fa.merge(me, on=['year', 'gvkey'])

    fa['be_me'] = fa.be / fa.me_dec
    fa = fa.set_index(['year', 'permno'])

    # SIZE
    size_col = 'me_jun'
    size_split = [0.5, 1.0]
    size_labels = ['s', 'b', 'smb']
    size_by = size_col + '_by'
    fa[size_by] = np.where((fa.exchcd == 1) & (fa.rcount > 1), fa[size_col], np.nan)  # NYSE and at least 2 years in funda
    fa['size_class'] = classify(fa[size_col], size_split, ascending=True, by_array=fa[size_by])

    # BM
    bm_col = 'be_me'
    bm_split = [0.3, 0.7, 1.0]
    bm_labels = ['h', 'm', 'l', 'hml']
    bm_by = bm_col + '_by'
    fa[bm_by] = np.where((fa.exchcd == 1) & (fa.rcount > 1), fa[bm_col], np.nan)  # NYSE and at least 2 years in funda
    fa['bm_class'] = classify(fa[bm_col], bm_split, ascending=False, by_array=fa[bm_by])

    # Remove unclassified rows. This happens if there are no NYSE stocks with at least 2 years in funda.
    fa = fa[fa['size_class'].isin([0, 1]) & fa['bm_class'].isin([0, 1, 2])]

    # Merge annual fa with monthly cm
    cm['year'] = add_months(cm.date, -18).dt.year
    cm = pd.merge(cm, fa[fa.columns.difference(cm.columns)], on=['permno', 'year'])
    cm = cm.set_index(['date', 'permno'])

    # Double sort on size and bm.
    ff_factors = two_dim_sort(cm, 'bm_class', 'size_class', 'ret', weight_col='me_1', output_dim=2)

    relabel_class(ff_factors, bm_labels, axis=0)
    relabel_class(ff_factors, size_labels, axis=1)
    ff_factors = ff_factors.unstack()
    ff_factors.columns = ff_factors.columns.map(''.join)

    # smb and hml factors
    ff_factors['hml'] = (ff_factors['shml'] + ff_factors['bhml']) / 2
    ff_factors['smb'] = (ff_factors['smbh'] + ff_factors['smbm'] + ff_factors['smbl']) / 3
    drop_columns(ff_factors, ['shml', 'bhml', 'smbh', 'smbm', 'smbl', 'smbhml'])

    # Firm count
    ff_nfirms = two_dim_sort(cm, 'bm_class', 'size_class', 'ret', function='count', add_long_short=False, output_dim=2)

    relabel_class(ff_nfirms, bm_labels, axis=0)
    relabel_class(ff_nfirms, size_labels, axis=1)
    ff_nfirms = ff_nfirms.unstack()
    ff_nfirms.columns = ff_nfirms.columns.map(''.join)

    ff_nfirms['hml'] = ff_nfirms['bh'] + ff_nfirms['sh'] + ff_nfirms['bl'] + ff_nfirms['sl']
    ff_nfirms['smb'] = ff_nfirms['sh'] + ff_nfirms['sm'] + ff_nfirms['sl'] + \
                       ff_nfirms['bh'] + ff_nfirms['bm'] + ff_nfirms['bl']

    return ff_factors, ff_nfirms



###############################################################################
#
# WRDS Code
#
###############################################################################
import datetime as dt
import wrds
from scipy import stats

def make_ff_factors_wrds():
    """
    This is a simple copy of the WRDS code.

    https://wrds-www.wharton.upenn.edu/pages/support/applications/python-replications/fama-french-factors-python/

    Returns:
        factors, number of firms.
    """
    ###################
    # Connect to WRDS #
    ###################
    conn = wrds.Connection()

    ###################
    # Compustat Block #
    ###################
    comp = conn.raw_sql("""
                        select gvkey, datadate, at, pstkl, txditc,
                        pstkrv, seq, pstk
                        from comp.funda
                        where indfmt='INDL' 
                        and datafmt='STD'
                        and popsrc='D'
                        and consol='C'
                        and datadate >= '01/01/1959'
                        """, date_cols=['datadate'])

    comp['year'] = comp['datadate'].dt.year

    # create preferrerd stock
    comp['ps'] = np.where(comp['pstkrv'].isnull(), comp['pstkl'], comp['pstkrv'])
    comp['ps'] = np.where(comp['ps'].isnull(), comp['pstk'], comp['ps'])
    comp['ps'] = np.where(comp['ps'].isnull(), 0, comp['ps'])
    comp['txditc'] = comp['txditc'].fillna(0)

    # create book equity
    comp['be']=comp['seq']+comp['txditc']-comp['ps']
    comp['be']=np.where(comp['be']>0, comp['be'], np.nan)

    # number of years in Compustat
    comp = comp.sort_values(by=['gvkey', 'datadate'])
    comp['count'] = comp.groupby(['gvkey']).cumcount()

    comp = comp[['gvkey', 'datadate', 'year', 'be', 'count']]

    ###################
    # CRSP Block      #
    ###################
    # sql similar to crspmerge macro
    crsp_m = conn.raw_sql("""
                          select a.permno, a.permco, a.date, b.shrcd, b.exchcd,
                          a.ret, a.retx, a.shrout, a.prc
                          from crsp.msf as a
                          left join crsp.msenames as b
                          on a.permno=b.permno
                          and b.namedt<=a.date
                          and a.date<=b.nameendt
                          where a.date between '01/01/1959' and '12/31/2017'
                          and b.exchcd between 1 and 3
                          """, date_cols=['date'])

    # change variable format to int
    crsp_m[['permco', 'permno', 'shrcd', 'exchcd']] = crsp_m[['permco', 'permno', 'shrcd', 'exchcd']].astype(int)

    # Line up date to be end of month
    crsp_m['jdate'] = crsp_m['date'] + MonthEnd(0)

    # add delisting return
    dlret = conn.raw_sql("""
                         select permno, dlret, dlstdt 
                         from crsp.msedelist
                         """, date_cols=['dlstdt'])

    dlret.permno = dlret.permno.astype(int)
    # dlret['dlstdt']=pd.to_datetime(dlret['dlstdt'])
    dlret['jdate'] = dlret['dlstdt'] + MonthEnd(0)

    crsp = pd.merge(crsp_m, dlret, how='left', on=['permno', 'jdate'])
    crsp['dlret'] = crsp['dlret'].fillna(0)
    crsp['ret'] = crsp['ret'].fillna(0)

    # retadj factors in the delisting returns
    crsp['retadj'] = (1 + crsp['ret']) * (1 + crsp['dlret']) - 1

    # calculate market equity
    crsp['me'] = crsp['prc'].abs() * crsp['shrout']
    crsp = crsp.drop(['dlret', 'dlstdt', 'prc', 'shrout'], axis=1)
    crsp = crsp.sort_values(by=['jdate', 'permco', 'me'])

    ### Aggregate Market Cap ###
    # sum of me across different permno belonging to same permco a given date
    crsp_summe = crsp.groupby(['jdate', 'permco'])['me'].sum().reset_index()

    # largest mktcap within a permco/date
    crsp_maxme = crsp.groupby(['jdate', 'permco'])['me'].max().reset_index()

    # join by jdate/maxme to find the permno
    crsp1 = pd.merge(crsp, crsp_maxme, how='inner', on=['jdate', 'permco', 'me'])

    # drop me column and replace with the sum me
    crsp1 = crsp1.drop(['me'], axis=1)

    # join with sum of me to get the correct market cap info
    crsp2 = pd.merge(crsp1, crsp_summe, how='inner', on=['jdate', 'permco'])

    # sort by permno and date and also drop duplicates
    crsp2 = crsp2.sort_values(by=['permno', 'jdate']).drop_duplicates()

    # keep December market cap
    crsp2['year'] = crsp2['jdate'].dt.year
    crsp2['month'] = crsp2['jdate'].dt.month
    decme = crsp2[crsp2['month'] == 12]
    decme = decme[['permno', 'date', 'jdate', 'me', 'year']].rename(columns={'me': 'dec_me'})

    ### July to June dates
    crsp2['ffdate'] = crsp2['jdate'] + MonthEnd(-6)
    crsp2['ffyear'] = crsp2['ffdate'].dt.year
    crsp2['ffmonth'] = crsp2['ffdate'].dt.month
    crsp2['1+retx'] = 1 + crsp2['retx']
    crsp2 = crsp2.sort_values(by=['permno', 'date'])

    # cumret by stock
    crsp2['cumretx'] = crsp2.groupby(['permno', 'ffyear'])['1+retx'].cumprod()

    # lag cumret
    crsp2['lcumretx'] = crsp2.groupby(['permno'])['cumretx'].shift(1)

    # lag market cap
    crsp2['lme'] = crsp2.groupby(['permno'])['me'].shift(1)

    # if first permno then use me/(1+retx) to replace the missing value
    crsp2['count'] = crsp2.groupby(['permno']).cumcount()
    crsp2['lme'] = np.where(crsp2['count'] == 0, crsp2['me'] / crsp2['1+retx'], crsp2['lme'])

    # baseline me
    mebase = crsp2[crsp2['ffmonth'] == 1][['permno', 'ffyear', 'lme']].rename(columns={'lme': 'mebase'})

    # merge result back together
    crsp3 = pd.merge(crsp2, mebase, how='left', on=['permno', 'ffyear'])
    crsp3['wt'] = np.where(crsp3['ffmonth'] == 1, crsp3['lme'], crsp3['mebase'] * crsp3['lcumretx'])

    decme['year'] = decme['year'] + 1
    decme = decme[['permno', 'year', 'dec_me']]

    # Info as of June
    crsp3_jun = crsp3[crsp3['month'] == 6]

    crsp_jun = pd.merge(crsp3_jun, decme, how='inner', on=['permno', 'year'])
    crsp_jun = crsp_jun[
        ['permno', 'date', 'jdate', 'shrcd', 'exchcd', 'retadj', 'me', 'wt', 'cumretx', 'mebase', 'lme', 'dec_me']]
    crsp_jun = crsp_jun.sort_values(by=['permno', 'jdate']).drop_duplicates()

    #######################
    # CCM Block           #
    #######################
    ccm = conn.raw_sql("""
                      select gvkey, lpermno as permno, linktype, linkprim, 
                      linkdt, linkenddt
                      from crsp.ccmxpf_linktable
                      where substr(linktype,1,1)='L'
                      and (linkprim ='C' or linkprim='P')
                      """, date_cols=['linkdt', 'linkenddt'])

    # if linkenddt is missing then set to today date
    ccm['linkenddt'] = ccm['linkenddt'].fillna(pd.to_datetime('today'))

    ccm1 = pd.merge(comp[['gvkey', 'datadate', 'be', 'count']], ccm, how='left', on=['gvkey'])
    ccm1['yearend'] = ccm1['datadate'] + pd.tseries.offsets.YearEnd(0)
    ccm1['jdate'] = ccm1['yearend'] + pd.tseries.offsets.MonthEnd(6)

    # set link date bounds
    ccm2 = ccm1[(ccm1['jdate'] >= ccm1['linkdt']) & (ccm1['jdate'] <= ccm1['linkenddt'])]
    ccm2 = ccm2[['gvkey', 'permno', 'datadate', 'yearend', 'jdate', 'be', 'count']]

    # link comp and crsp
    ccm_jun = pd.merge(crsp_jun, ccm2, how='inner', on=['permno', 'jdate'])
    ccm_jun['beme'] = ccm_jun['be'] * 1000 / ccm_jun['dec_me']

    # select NYSE stocks for bucket breakdown
    # exchcd = 1 and positive beme and positive me and shrcd in (10,11) and at least 2 years in comp
    nyse = ccm_jun[(ccm_jun['exchcd'] == 1) & (ccm_jun['beme'] > 0) & (ccm_jun['me'] > 0) & \
                   (ccm_jun['count'] >= 1) & ((ccm_jun['shrcd'] == 10) | (ccm_jun['shrcd'] == 11))]

    # size breakdown
    nyse_sz = nyse.groupby(['jdate'])['me'].median().to_frame().reset_index().rename(columns={'me': 'sizemedn'})

    # beme breakdown
    nyse_bm = nyse.groupby(['jdate'])['beme'].describe(percentiles=[0.3, 0.7]).reset_index()
    nyse_bm = nyse_bm[['jdate', '30%', '70%']].rename(columns={'30%': 'bm30', '70%': 'bm70'})

    nyse_breaks = pd.merge(nyse_sz, nyse_bm, how='inner', on=['jdate'])

    # join back size and beme breakdown
    ccm1_jun = pd.merge(ccm_jun, nyse_breaks, how='left', on=['jdate'])

    # function to assign sz and bm bucket
    def sz_bucket(row):
        if row['me'] == np.nan:
            value = ''
        elif row['me'] <= row['sizemedn']:
            value = 'S'
        else:
            value = 'B'
        return value

    def bm_bucket(row):
        if 0 <= row['beme'] <= row['bm30']:
            value = 'L'
        elif row['beme'] <= row['bm70']:
            value = 'M'
        elif row['beme'] > row['bm70']:
            value = 'H'
        else:
            value = ''
        return value

    # assign size portfolio
    ccm1_jun['szport'] = np.where((ccm1_jun['beme'] > 0) & (ccm1_jun['me'] > 0) & (ccm1_jun['count'] >= 1),
                                  ccm1_jun.apply(sz_bucket, axis=1), '')

    # assign book-to-market portfolio
    ccm1_jun['bmport'] = np.where((ccm1_jun['beme'] > 0) & (ccm1_jun['me'] > 0) & (ccm1_jun['count'] >= 1),
                                  ccm1_jun.apply(bm_bucket, axis=1), '')

    # create positivebmeme and nonmissport variable
    ccm1_jun['posbm'] = np.where((ccm1_jun['beme'] > 0) & (ccm1_jun['me'] > 0) & (ccm1_jun['count'] >= 1), 1, 0)
    ccm1_jun['nonmissport'] = np.where((ccm1_jun['bmport'] != ''), 1, 0)

    # store portfolio assignment as of June
    june = ccm1_jun[['permno', 'date', 'jdate', 'bmport', 'szport', 'posbm', 'nonmissport']]
    june['ffyear'] = june['jdate'].dt.year

    # merge back with monthly records
    crsp3 = crsp3[['date', 'permno', 'shrcd', 'exchcd', 'retadj', 'me', 'wt', 'cumretx', 'ffyear', 'jdate']]
    ccm3 = pd.merge(crsp3,
                    june[['permno', 'ffyear', 'szport', 'bmport', 'posbm', 'nonmissport']], how='left',
                    on=['permno', 'ffyear'])

    # keeping only records that meet the criteria
    ccm4 = ccm3[(ccm3['wt'] > 0) & (ccm3['posbm'] == 1) & (ccm3['nonmissport'] == 1) &
                ((ccm3['shrcd'] == 10) | (ccm3['shrcd'] == 11))]

    ############################
    # Form Fama French Factors #
    ############################

    # function to calculate value weighted return
    def wavg(group, avg_name, weight_name):
        d = group[avg_name]
        w = group[weight_name]
        try:
            return (d * w).sum() / w.sum()
        except ZeroDivisionError:
            return np.nan

    # value-weigthed return
    vwret = ccm4.groupby(['jdate', 'szport', 'bmport']).apply(wavg, 'retadj', 'wt').to_frame().reset_index().rename(
        columns={0: 'vwret'})
    vwret['sbport'] = vwret['szport'] + vwret['bmport']

    # firm count
    vwret_n = ccm4.groupby(['jdate', 'szport', 'bmport'])['retadj'].count().reset_index().rename(
        columns={'retadj': 'n_firms'})
    vwret_n['sbport'] = vwret_n['szport'] + vwret_n['bmport']

    # tranpose
    ff_factors = vwret.pivot(index='jdate', columns='sbport', values='vwret').reset_index()
    ff_nfirms = vwret_n.pivot(index='jdate', columns='sbport', values='n_firms').reset_index()
    # create SMB and HML factors
    ff_factors['WH'] = (ff_factors['BH'] + ff_factors['SH']) / 2
    ff_factors['WL'] = (ff_factors['BL'] + ff_factors['SL']) / 2
    ff_factors['WHML'] = ff_factors['WH'] - ff_factors['WL']

    ff_factors['WB'] = (ff_factors['BL'] + ff_factors['BM'] + ff_factors['BH']) / 3
    ff_factors['WS'] = (ff_factors['SL'] + ff_factors['SM'] + ff_factors['SH']) / 3
    ff_factors['WSMB'] = ff_factors['WS'] - ff_factors['WB']
    ff_factors = ff_factors.rename(columns={'jdate': 'date'})

    # n firm count
    ff_nfirms['H'] = ff_nfirms['SH'] + ff_nfirms['BH']
    ff_nfirms['L'] = ff_nfirms['SL'] + ff_nfirms['BL']
    ff_nfirms['HML'] = ff_nfirms['H'] + ff_nfirms['L']

    ff_nfirms['B'] = ff_nfirms['BL'] + ff_nfirms['BM'] + ff_nfirms['BH']
    ff_nfirms['S'] = ff_nfirms['SL'] + ff_nfirms['SM'] + ff_nfirms['SH']
    ff_nfirms['SMB'] = ff_nfirms['B'] + ff_nfirms['S']
    ff_nfirms['TOTAL'] = ff_nfirms['SMB']
    ff_nfirms = ff_nfirms.rename(columns={'jdate': 'date'})

    ###################
    # Compare With FF #
    ###################
    _ff = conn.get_table(library='ff', table='factors_monthly')
    _ff = _ff[['date', 'smb', 'hml']]
    _ff['date'] = _ff['date'] + MonthEnd(0)

    _ffcomp = pd.merge(_ff, ff_factors[['date', 'WSMB', 'WHML']], how='inner', on=['date'])
    _ffcomp70 = _ffcomp[_ffcomp['date'] >= '01/01/1970']
    print(stats.pearsonr(_ffcomp70['smb'], _ffcomp70['WSMB']))
    print(stats.pearsonr(_ffcomp70['hml'], _ffcomp70['WHML']))

    plt.figure(figsize=(16, 12))
    plt.suptitle('Comparison of Results', fontsize=20)

    ax1 = plt.subplot(211)
    ax1.set_title('SMB', fontsize=15)
    ax1.set_xlim([dt.datetime(1962, 6, 1), dt.datetime(2017, 12, 31)])
    ax1.plot(_ffcomp['smb'], 'r--', _ffcomp['WSMB'], 'b-')
    ax1.legend(('smb', 'WSMB'), loc='upper right', shadow=True)

    ax2 = plt.subplot(212)
    ax2.set_title('HML', fontsize=15)
    ax2.plot(_ffcomp['hml'], 'r--', _ffcomp['WHML'], 'b-')
    ax2.set_xlim([dt.datetime(1962, 6, 1), dt.datetime(2017, 12, 31)])
    ax2.legend(('hml', 'WHML'), loc='upper right', shadow=True)

    plt.subplots_adjust(top=0.92, hspace=0.2)

    plt.show()

    ff_factors.set_index('date', inplace=True)
    ff_factors = ff_factors.rename(
        columns={'BH': 'bh', 'BL': 'bl', 'BM': 'bm', 'SH': 'sh', 'SL': 'sl', 'SM': 'sm', 'WH': 'h', 'WL': 'l',
                 'WHML': 'hml', 'WB': 'b', 'WS': 's', 'WSMB': 'smb'})

    ff_nfirms.set_index('date', inplace=True)
    ff_nfirms = ff_nfirms.rename(
        columns={'BH': 'bh', 'BL': 'bl', 'BM': 'bm', 'SH': 'sh', 'SL': 'sl', 'SM': 'sm', 'H': 'h', 'L': 'l',
                 'HML': 'hml', 'B': 'b', 'S': 's', 'SMB': 'smb'})


    return ff_factors, ff_nfirms


if __name__ == '__main__':
    os.chdir('../')

    ##############################
    # Generate factors
    ##############################
    # ff factors from pyanomaly
    ff_factors1, ff_nfirms1 = make_ff_factors()

    # ff factors from WRDS code
    ff_factors2, ff_nfirms2 = make_ff_factors_wrds()

    # ff factors from French website
    ff_factors3 = WRDS.read_data('factors_monthly')
    ff_factors3['date'] = to_month_end(ff_factors3.dateff)
    ff_factors3.set_index('date', inplace=True)

    ##############################
    # Compare results
    ##############################
    # To compare from 1970.
    # ff_factors1 = ff_factors1['1970':]
    # ff_factors2 = ff_factors2['1970':]
    # ff_factors3 = ff_factors3['1970':]
    # ff_nfirms1 = ff_nfirms1['1970':]
    # ff_nfirms2 = ff_nfirms2['1970':]

    print('\nCompare pyanomaly with WRDS')
    print('factors')
    ff_factors12 = compare_data(ff_factors1, ff_factors2, on='date', tolerance=0.1)
    print('\nnumber of firms')
    ff_nfirms12 = compare_data(ff_nfirms1, ff_nfirms2, on='date', tolerance=0.1)
    ff_factors12[['smb_x', 'smb_y']].plot()
    plt.show()

    print('\nCompare pyanomaly with FF')
    ff_factors13 = compare_data(ff_factors1, ff_factors3, on='date', tolerance=0.1)
    ff_factors13[['smb_x', 'smb_y']].plot()
    plt.show()

    print('\nCompare WRDS with FF')
    ff_factors23 = compare_data(ff_factors2, ff_factors3, on='date', tolerance=0.1)
    ff_factors23[['smb_x', 'smb_y']].plot()
    plt.show()



import time as t
import sys
import os
import pandas as pd

log_path = None


def set_log_path(fpath):
    global log_path
    log_path = fpath

    fdir = fpath[:max(fpath.rfind('/'), fpath.rfind('\\'))]
    if not os.path.isdir(fdir):
        os.mkdir(fdir)


def log(msg, headmsg=None, header=True):
    if isinstance(msg, pd.DataFrame) or isinstance(msg, pd.Series):
        msg = f'\n{msg}'

    if header and headmsg:
        head = f"[{t.strftime('%Y/%m/%d %H:%M', t.localtime())}: {headmsg}] "
    elif header:
        head = f"[{t.strftime('%Y/%m/%d %H:%M', t.localtime())}] "
    elif headmsg:
        head = f"[{headmsg}] "
    else:
        head = ""

    text = head + str(msg)
    print(text)
    if log_path and (log_path != sys.stdout):
        with open(log_path, 'a') as f:
            print(text, file=f)


def err(msg):
    return log(msg, 'ERROR')


def warn(msg):
    return log(msg, 'WARNING')


def drawline(level=0, width=80):
    if level == 0:
        log("#" * width)
    elif level == 1:
        log("*" * width)
    elif level == 2:
        log("-" * width)
    elif isinstance(level, str):
        log(level * width)


def start_timer(msg='', test_id=None):
    global tstart, tprev, testid
    tstart = t.perf_counter()
    tprev = tstart
    testid = test_id
    log(f"Timer On: {msg}", testid)


def elapsed_time(msg=''):
    try:
        global tstart, tprev, testid
        tnow = t.perf_counter()
        m, s = divmod(int(tnow - tstart + 1), 60)
        h, m = divmod(m, 60)
        log(f"Elapsed [{h:d}:{m:02d}:{s:02d}, {tnow - tprev:0.3f}]: {msg}", testid)
        tprev = tnow
    except:
        start_timer(msg)



# -*- coding: utf-8 -*-

"""
Python library for asset pricing research.

"""

__title__ = 'pyanomaly'
__version__ = '0.9'
__author__ = 'Chulwoo Han and Jongho Kang'
__copyright__ = '2021 @ Chulwoo Han and Jongho Kang'

from pyanomaly.globals import *
from pyanomaly.fileio import *



import shutil
import pandas as pd
import datetime as dt
import sqlite3
import wrds
import asyncio

from pyanomaly.globals import *
from pyanomaly.datatools import to_month_end
from pyanomaly.datatools import inspect_data

################################################################################
#
# Class to download/handle WRDS data.
#
# Useful links
#  - CRSP overview
#     https://wrds-www.wharton.upenn.edu/pages/support/data-overview/wrds-overview-crsp-us-stock-database/
#  - CRSP-COMPUSTST merge overview
#     https://wrds-www.wharton.upenn.edu/pages/support/manuals-and-overviews/crsp/crspcompustat-merged-ccm/wrds-overview-crspcompustat-merged-ccm/
#  - CRSP annual update table list
#     https://wrds-www.wharton.upenn.edu/data-dictionary/crsp_a_indexes/
#
# Exchcd
# -2: Halted by the NYSE or AMEX
# -1: Suspended by the NYSE, AMEX, or NASDAQ
#  0: Not Trading on NYSE, AMEX, or NASDAQ
#  1: New York Stock Exchange
#  2: American Stock Exchange
#  3: NASDAQ
#
# Shrcd (https://wrds-www.wharton.upenn.edu/data-dictionary/form_metadata/crsp_a_stock_msf_identifyinginformation/shrcd/)
#  1st digit
#  1: Ordinary Common Shares
#  2: Certificates
#  3: ADRs (American Depository Receipts)
#  4: SBIs (Shares of Beneficial Interest)
#  5: Units (Depository Units, Units of Beneficial Interest, Units of Limited Partnership Interest, Depository Receipts, etc.)
#  2nd digit
#  0: Securities which have not been further defined.
#  1: Securities which need not be further defined.
#  2: Companies incorporated outside the US
#  3: Americus Trust Components (Primes and Scores).
#  4: Closed-end funds.
#  5: Closed-end fund companies incorporated outside the US
#  8: REIT's (Real Estate Investment Trusts).
################################################################################


class WRDS():
    """Class to download/handle WRDS data.

    Args:
        wrds_username: WRDS username. Necessary only when downloading data: can be set to None when reading
            data from files.

    **Methods for Data Download**

        * ``download_table()``
        * ``download_table_async()``
        * ``download_funda()``
        * ``download_fundq()``
        * ``download_sf()``
        * ``download_seall()``
        * ``download_secd()``
        * ``download_g_secd()``
        * ``download_all()``

    **Other Methods**

        * ``create_pgpass_file()``
        * ``merge_sf_with_seall()``
        * ``add_gvkey_to_crsp()``
        * ``preprocess_crsp()``
        * ``convert_fund_currency_to_usd()``
        * ``get_risk_free_rate()``
        * ``add_gvkey_to_crsp()``
        * ``save_data()``
        * ``read_data()``
        * ``save_as_csv()``

    References:
        CRSP overview: https://wrds-www.wharton.upenn.edu/pages/support/data-overview/wrds-overview-crsp-us-stock-database/

        CRSP-Compustat merge: https://wrds-www.wharton.upenn.edu/pages/support/manuals-and-overviews/crsp/crspcompustat-merged-ccm/wrds-overview-crspcompustat-merged-ccm/

        CRSP annual update tables: https://wrds-www.wharton.upenn.edu/data-dictionary/crsp_a_indexes/

        shrcd: https://wrds-www.wharton.upenn.edu/data-dictionary/form_metadata/crsp_a_stock_msf_identifyinginformation/shrcd/

        exchcd: https://wrds-www.wharton.upenn.edu/data-dictionary/form_metadata/crsp_a_stock_msf_identifyinginformation/exchcd/

    Attributes:
        db: A wrds object to connect to WRDS database.
    """
    def __init__(self, wrds_username=None):
        if wrds_username:
            self.db = wrds.Connection(wrds_username=wrds_username)

    def create_pgpass_file(self):
        """Create pgpass file. Need to be called only once. Once created, you don't need to enter password when
        connecting to WRDS.
        """

        self.db.create_pgpass_file()

    ################################################################################
    #
    # DATA DOWNLOAD
    #
    ################################################################################
    def download_table(self, library, table, obs=-1, offset=0, columns=None, coerce_float=None, index_col=None,
                       date_cols=None):
        """Download a table from WRDS library.

        This is a wrapping function of ``wrds.get_table()``.
        The queried table is saved in config.input_dir/library/table.pickle. Useful when downloading an entire table.

        Args:
            library: WRDS library. e.g., crsp, comp, ...
            table: A table in `library`.
            obs: See ``wrds.get_table()``.
            offset: See ``wrds.get_table()``.
            columns: See ``wrds.get_table()``.
            coerce_float: See ``wrds.get_table()``.
            index_col: See ``wrds.get_table()``.
            date_cols: See ``wrds.get_table()``.
        """

        elapsed_time(f'Downloading {library}.{table}...')
        data = self.db.get_table(library, table, obs, offset, columns, coerce_float, index_col, date_cols)
        data = self._convert_types(data, [(library, table)])

        self.save_data(data, table, library)
        elapsed_time(f'Download complete. Size: {data.shape}')

    def _convert_types(self, data, src_tables, type=float, date_cols=None):
        """Convert dtype of double precision fields to float.

        The dtype of a double precision field is object when first
        downloaded if the field contains None, which can significantly slower computations.

        Args:
            data: Dataframe of which dtypes to be converted.
            src_tables: list of (library, table) that are used to create the data.
            type: float or np.float32. Numeric fields will be converted to this type. For a large dataset, converting
            to float32 can save disc space and read/write time.
            date_cols: list of date columns. The dtype of these columns will be converted to pd.Timestamp.

        Returns:
            data: type-converted data.
        """

        desc = [self.db.describe_table(*src) for src in src_tables]  # get table descriptions
        desc = pd.concat(desc).drop_duplicates(subset=['name'])

        for col in data:
            if str(desc.loc[desc.name == col, 'type'].values[0]) == 'DOUBLE PRECISION':
                data[col] = data[col].astype(type, errors='ignore')
            if (date_cols is not None) and (col in date_cols):
                data[col] = pd.to_datetime(data[col])

        return data

    def _generate_sql(self, library, table, sql, date_col):
        if not sql or ('from' not in sql.lower()):
            select = sql or '*'

            sql = f"""
                SELECT {select}
                FROM 
                    {library}.{table}
                WHERE {date_col} between '{{}}' and '{{}}'
                """
        print(sql)
        return sql

    def _download_table_async(self, table, sql, sdate, edate):
        log(f'Downloading... {table}: {sdate}-{edate}')
        # await asyncio.sleep(.01)

        data = self.db.raw_sql(sql.format(sdate, edate))
        if data.empty:
            return

        self.save_data(data, f"{table}_{sdate.replace('-', '')}_{edate.replace('-', '')}", 'tmp')
        log(f"Download complete. {table}: {sdate}-{edate}")

    def download_table_async(self, library, table, sql=None, date_col=None, sdate=None, edate=None, interval=5, src_tables=None, type=float, date_cols=None, run_in_executer=True):
        """Asynchronous download.

        This method splits the total period into `interval` years and downloads data for each sub-period
        asynchronously. If download fails, it can be started from the failed date: already downloaded files will be
        gathered together. This method allow us to download a large table, e.g., crsp.dsf, reliably without connection
        timeout and consumes much less memory than ``WRDS.download_table()``.

        NOTE:
            For a small table, this can be slower than ``WRDS.download_table()``.
            Table should have a date field (`date_col`) to split the period.

        Args:
            library: WRDS library.
            table: WRDS table. If a complete query is given, this can be any name: this is used only as the file name
                for the data.
            sql: String of the fields to select or a complete query statement. See below.
            date_col: Date field on which downloaing will be split.
            sdate: Start date ('yyyy-mm-dd'). If None, '1900-01-01'.
            edate: End date. If None, today.
            interval: Sub-period size in years.
            src_tables: List of (library, table) that are used in the query. `src_tables` are used to get data types
                and convert the types of double precision fields to float. When data is selected from a single table,
                `library.table`, this can be set to None.
            type: Type for numeric fields: float or np.float32. For a large dataset, converting to float32 can save
                disc space and read/write time.
            date_cols: List of date columns. The dtype of these columns will be converted to datetime.
            run_in_executer: If True, download data concurrently using `executer`. Setting this to True will increase
                download speed but can take up much memory.

        NOTE:
            * To download library.table, `library`, `table`, and `date_col` should be given.
            * To download selected fields from library.table, library, `table`, `sql`, and `date_col` should be given, where
              `sql` is a string of the fields to select, e.g.,

              >>> wrds = WRDS('user_name')
              >>> sql = 'permno, prc, exchcd'
              >>> wrds.download_table_async('crsp', 'msf', sql, 'date')

            * To download data using a complete query, `library`, `table`, and `sql` should be given, where `table` is a table
              name for the data (can be any name), and `sql` is a query statement. The `sql` should contain
              'WHERE [`date_col`] BETWEEN {} and {}'. See, e.g., the code of ``WRDS.download_sf()``.
        """

        elapsed_time(f'Asynchronous download: {table}')

        # Make a list of sub-periods.
        ldate = edate or str(dt.datetime.now())[:10]  # last date
        sdate = sdate or '1900-01-01'  # sub-period start date
        edate = min(str(pd.Timestamp(sdate) + pd.DateOffset(years=interval, days=-1))[:10], ldate)  # sub-period end date

        dates = []
        while sdate < ldate:
            dates.append((sdate, edate))
            sdate = str(int(sdate[:4]) + interval) + sdate[4:]
            edate = min(str(int(edate[:4]) + interval) + edate[4:], ldate)

        # Generate query statement.
        sql = self._generate_sql(library, table, sql, date_col)

        # Download and save data for each sub-period.
        if (date_col is None) and (sql and '{}' not in sql):  # Download at once.
            data = self.db.raw_sql(sql)
        else:
            if run_in_executer:
                loop = asyncio.get_event_loop()
                # aws = [self._download_table_async(table, sql, sdate, edate) for sdate, edate in dates]
                aws = [loop.run_in_executor(None, self._download_table_async, table, sql, sdate, edate) for sdate, edate in dates]
                loop.run_until_complete(asyncio.gather(*aws))
            else:
                for sdate, edate in dates:
                    self._download_table_async(table, sql, sdate, edate)


            # Read saved data files and save them as one big file.
            log('Gathering files...')
            tmp_dir = config.input_dir + 'tmp'  # sub-period files are temporarily saved here.
            data = []
            for f in os.listdir(tmp_dir):
                fpath = os.path.join(tmp_dir, f)
                if os.path.isfile(fpath) and (f[:len(table)] == table):
                    data.append(pd.read_pickle(fpath))
            data = pd.concat(data)

            for file in os.scandir(tmp_dir):
                print(file)
                os.remove(file.path)

        # type conversion
        if src_tables is None:
            src_tables = [(library, table)]
        data = self._convert_types(data, src_tables, type, date_cols)

        self.save_data(data, table, library)
        elapsed_time(f'Download complete: {table}')

    def download_sf(self, sdate=None, edate=None, monthly=True, run_in_executer=True):
        """ Download crsp.m(d)sf joined with crsp.m(d)senames.

        Args:
            sdate: Start date. e.g., '2000-01-01'. Set to None to download all data.
            edate: End date. Set to None to download all data.
            monthly: If True download msf else dsf.
            run_in_executer: If True, download concurrently. Faster but memory hungrier.
        """

        if monthly:
            fields = 'a.*, b.shrcd, b.exchcd, b.siccd'
            sf = 'msf'
            senames = 'msenames'
        else:  # for dsf, download only necessary columns to reduce file size.
            fields = 'a.permno, a.permco, a.date, bidlo, askhi, prc, ret, vol, shrout, cfacpr, cfacshr, b.shrcd, b.exchcd'
            sf = 'dsf'
            senames = 'dsenames'

        sql = f"""
            SELECT {fields}
            FROM crsp.{sf} as a
            LEFT JOIN crsp.{senames} as b
            ON a.permno = b.permno
            AND b.namedt <= a.date
            AND a.date <= b.nameendt
            WHERE a.date between '{{}}' and '{{}}'
            ORDER BY a.permno, a.date
        """

        src_tables = [('crsp', sf), ('crsp', senames)]
        type = float if monthly else np.float32  # For dsf, use float32 to save space.
        self.download_table_async('crsp', sf, sql, sdate=sdate, edate=edate, src_tables=src_tables, type=type, date_cols=['date'], run_in_executer=run_in_executer)

    def download_seall(self, sdate=None, edate=None, monthly=True, run_in_executer=True):
        """Download delist info from m(d)seall.

        Delist can be obtained from either mseall or msedelist. We use mseall since it contains exchcd, which is used
        when replacing missing dlret.
        The shrcd and exchcd in mseall are usually those before halting/suspension. If a stock in NYSE is halted, exchcd
        in msenames can be -2, whereas that in mseall is 1.

        Args:
            sdate: Start date. e.g., '2000-01-01'. Set to None to download all data.
            edate: End date. Set to None to download all data.
            monthly: If True download mseall else dseall.
            run_in_executer: If True, download concurrently. Faster but memory hungrier.
        """

        seall = 'mseall' if monthly else 'dseall'

        sql = f"""
            SELECT distinct permno, date, 
            dlret, dlstcd, shrcd, exchcd, 
            distcd, divamt  
            FROM crsp.{seall}
            WHERE date between '{{}}' and '{{}}'
        """

        type = float if monthly else np.float32
        self.download_table_async('crsp', seall, sql, sdate=sdate, edate=edate, type=type, date_cols=['date'], run_in_executer=run_in_executer)

    def download_funda(self, sdate=None, edate=None, run_in_executer=True):
        """Download comp.funda.

        Args:
            sdate: Start date. e.g., '2000-01-01'. Set to None to download all data.
            edate: End date. Set to None to download all data.
            run_in_executer: If True, download concurrently. Faster but memory hungrier.
        """

        sql = f"""
            SELECT 
                cusip, c.gvkey, datadate, c.cik, sic, naics, fyear,
                sale, revt, cogs, xsga, dp, xrd, xad, ib, gp, ebitda, ebit, nopi, spi, pi, txp, ni, txfed, txfo, txt, xint,
                capx, oancf, dvt, ob, gdwlia, gdwlip, gwo,
                rect, act, che, ppegt, invt, at, aco, intan, ao, ppent, gdwl, fatb, fatl,
                lct, dlc, dltt, lt, dm, dcvt, cshrc, dcpstk, pstk, ap, lco, lo, drc, drlt, txdi,
                ceq, scstkc, emp, csho, prcc_f,
                oibdp, oiadp, mii, xopr, xlr, xi, f.do, xido, ibc, dpc, xidoc,
                fincf, fiao, txbcof, ds, dltr, dlcch, prstkc, sstk, dv, dvc,
                ivao, ivst, re, txditc, txdb, itcb, seq, pstkrv, pstkl, mib, icapt, dltis, ajex, ppenb, ppenls, 
                curcd
                /* wcapt, ltdch, purtshr */
            FROM 
                comp.company as c,
                comp.funda as f
            WHERE c.gvkey = f.gvkey
            AND f.indfmt = 'INDL' AND f.datafmt = 'STD' AND f.popsrc = 'D' AND f.consol = 'C'
            AND datadate between '{{}}' and '{{}}'
            ORDER BY c.gvkey, datadate
            """

        src_tables = [('comp', 'funda'), ('comp', 'company')]
        self.download_table_async('comp', 'funda', sql, sdate=sdate, edate=edate, src_tables=src_tables, date_cols=['datadate'], run_in_executer=run_in_executer)

    def download_fundq(self, sdate=None, edate=None, run_in_executer=True):
        """Download comp.fundq.

        Args:
            sdate: Start date. e.g., '2000-01-01'. Set to None to download all data.
            edate: End date. Set to None to download all data.
            run_in_executer: If True, download concurrently. Faster but memory hungrier.
        """

        sql = f"""
            SELECT 
                cusip, c.gvkey, datadate, c.cik, sic, naics, fyr, fyearq, fqtr,
                saleq, revtq, cogsq, xsgaq, dpq, xrdq, ibq, nopiq, spiq, piq, txpq, niq, txtq, xintq,
                gdwliaq, gdwlipq,
                rectq, actq, cheq, ppegtq, invtq, atq, acoq, intanq, aoq, ppentq, gdwlq,
                lctq, dlcq, dlttq, ltq, pstkq, apq, lcoq, loq, drcq, drltq, txdiq,
                ceqq, cshoq, prccq, rdq, pstkrq, aqaq, aqdq, aqepsq, aqpq, arcedq,
                arceepsq, arceq, cibegniq, cicurrq, ciderglq, 
                oibdpq, oiadpq, miiq, xoprq, xiq, doq, xidoq,
                ivaoq, ivstq, req, txditcq, txdbq, seqq, mibq, icaptq, ajexq, cshprq, epspxq, 
                saley, revty, cogsy, xsgay,
                dpy, xrdy, iby, nopiy, spiy, piy, niy, txty, xinty, capxy, oancfy, gdwliay, gdwlipy,
                txdiy, scstkcy, oibdpy, oiadpy, miiy, xopry, xiy, doy, xidoy, ibcy, dpcy, xidocy, 
                fincfy, fiaoy, txbcofy, dltry, dlcchy, prstkcy, sstky, dvy, dltisy
                afudciy, amcy, aolochy, apalchy, aqay, aqcy, aqdy, aqepsy, aqpy, arcedy, 
                arceepsy, arcey, cdvcy, chechy, cibegniy, cicurry, cidergly, wcapq, 
                curcdq
            FROM 
                comp.company as c,
                comp.fundq as f
            WHERE c.gvkey = f.gvkey
            AND f.indfmt = 'INDL' AND f.datafmt = 'STD' AND f.popsrc = 'D' AND f.consol = 'C'
            AND datadate between '{{}}' and '{{}}'
            ORDER BY c.gvkey, datadate
            """

        src_tables = [('comp', 'fundq'), ('comp', 'company')]
        self.download_table_async('comp', 'fundq', sql, sdate=sdate, edate=edate, src_tables=src_tables, date_cols=['datadate'], run_in_executer=run_in_executer)

    def download_secd(self, sdate=None, edate=None, run_in_executer=True):
        """Download comp.secd.

        Args:
            sdate: Start date. e.g., '2000-01-01'. Set to None to download all data.
            edate: End date. Set to None to download all data.
            run_in_executer: If True, download concurrently. Faster but memory hungrier.
        """
        sql = 'gvkey, datadate, prccd, ajexdi, cshoc, iid'
        self.download_table_async('comp', 'secd', sql, 'datadate', sdate, edate, run_in_executer=run_in_executer)

    def download_g_secd(self, sdate=None, edate=None, run_in_executer=True):
        """Download comp.g_secd.

        Args:
            sdate: Start date. e.g., '2000-01-01'. Set to None to download all data.
            edate: End date. Set to None to download all data.
            run_in_executer: If True, download concurrently. Faster but memory hungrier.
        """
        sql = 'gvkey, datadate, prccd, ajexdi, cshoc, iid'
        self.download_table_async('comp', 'g_secd', sql, 'datadate', sdate, edate, run_in_executer=run_in_executer)

    def download_all(self, run_in_executer=True):
        """Download all tables.

        Currently, this function downloads the following tables:

        * comp.funda
        * comp.fundq
        * crsp.msf (merged with crsp.msenames)
        * crsp.dsf (merged with crsp.dsenames)
        * crsp.mseall
        * crsp.dseall
        * crsp.ccmxpf_linktable
        * crsp.mcti
        * ff.factors_monthly
        * ff.factors_daily
        * comp.exrt_dly

        Args:
            run_in_executer: If True, download concurrently. Faster but memory hungrier.
        """

        self.download_funda(run_in_executer=run_in_executer)
        self.download_fundq(run_in_executer=run_in_executer)
        self.download_sf(monthly=True, run_in_executer=run_in_executer)
        self.download_sf(monthly=False, run_in_executer=run_in_executer)
        self.download_seall(monthly=True, run_in_executer=run_in_executer)
        self.download_seall(monthly=False, run_in_executer=run_in_executer)

        self.download_table('crsp', 'ccmxpf_linktable', date_cols=['linkdt', 'linkenddt'])
        self.download_table('crsp', 'mcti', date_cols=['caldt'])
        self.download_table('ff', 'factors_monthly', date_cols=['date', 'dateff'])
        self.download_table('ff', 'factors_daily', date_cols=['date'])
        self.download_table('comp', 'exrt_dly', date_cols=['datadate'])

    ################################################################################
    #
    # UTILITIES
    #
    ################################################################################
    @staticmethod
    def merge_sf_with_seall(sf, monthly=True, fill_method=1):
        """Merge m(d)sf with m(d)seall.

        This method adjusts m(d)sf return ('ret') using m(d)seall delist return ('dlret'). The adjusted return replaces
        'ret' and 'dlret' column is added to m(d)sf.
        For msf, this method also adds cash dividend columns, 'cash_div', to m(d)sf.

        Args:
            sf: m(d)sf DataFrame.
            monthly: `sf` = msf if True, else `sf` = dsf.
            fill_method: Method to fill missing dlret. 0: don't fill, 1: JKP code, or 2: GHZ code.

                - fill_method = 1:

                    - `dlret` = -0.30 if `dlstcd` is 500 or between 520 and 584.
                - fill_method = 2:

                    - `dlret` = -0.35 if `dlstcd` is 500 or between 520 and 584, and `exchcd` is 1 or 2.
                    - `dlret` = -0.55 if `dlstcd` is 500 or between 520 and 584, and `exchcd` is 3.

        Returns:
            m(d)sf with adjusted return (and cash dividend).

        References:
            Delist codes: http://www.crsp.com/products/documentation/delisting-codes

        """

        f = 'm' if monthly else 'd'
        elapsed_time(f'Merging {f}sf with {f}seall...')

        seall = WRDS.read_data(f'{f}seall')

        #######################################
        # DLRET
        #######################################
        delist = seall[['permno', 'date', 'dlret', 'dlstcd', 'shrcd', 'exchcd']].dropna(subset=['dlstcd']).drop_duplicates().copy()
        log(f'{f}sf shape: {sf.shape}, delist{f} shape: {delist.shape}')

        # Fill missing delist return
        log(f'Missing dlret: {delist.dlret.isna().sum()}')
        if fill_method == 1:  # from JKP SAS code
            ridx = delist.dlret.isnull() & ((delist.dlstcd == 500) | ((delist.dlstcd >= 520) & (delist.dlstcd <= 584)))
            delist.loc[ridx, 'dlret'] = -0.30
        elif fill_method == 2:  # from GHZ SAS code
            ridx = delist.dlret.isnull() & ((delist.dlstcd == 500) | ((delist.dlstcd >= 520) & (delist.dlstcd <= 584)))
            delist.loc[ridx & ((delist.exchcd == 1) | (delist.exchcd == 2)), 'dlret'] = -0.35
            delist.loc[ridx & (delist.exchcd == 3), 'dlret'] = -0.55
        else:  # do nothing
            pass

        log(f'Missing dlret after filling: {delist.dlret.isna().sum()}')

        # Merge sf with delist
        sf = pd.merge(sf, delist[['permno', 'date', 'dlret']], on=['permno', 'date'], how='left')

        is_ret_nan = sf.ret.isna() & sf.dlret.isna()
        sf.ret = (1 + sf.ret.fillna(0)) * (1 + sf.dlret.fillna(0)) - 1
        sf.loc[is_ret_nan, 'ret'] = np.nan  # Set ret = None if both ret and dlret missing.

        if not monthly:  # if dsf, stop here: no need to add cash dividend.
            elapsed_time(f'{f}sf and {f}seall merged. {f}sf shape: {sf.shape}')
            return sf

        #######################################
        # CASH DIST
        #######################################
        # Used in divi, divo.
        # Be careful there can be multiple distcd's for a (permno, date) pair.
        dist = seall[['permno', 'date', 'distcd', 'divamt']].dropna(subset=['distcd']).drop_duplicates().copy()
        dist['is_cash_div'] = (dist.distcd % 1000 // 100).isin([2, 3])  # cash if 2nd digit in (2, 3)
        dist.loc[dist.is_cash_div, 'cash_div'] = dist.loc[dist.is_cash_div, 'divamt']
        dist = dist.groupby(['permno', 'date'])['cash_div'].sum()
        # if not monthly:
        #     dist['cash_div'] = dist['cash_div'].astype('float32')
        sf = pd.merge(sf, dist, on=['permno', 'date'], how='left')

        elapsed_time(f'{f}sf and {f}seall merged. {f}sf shape: {sf.shape}')
        return sf

    @staticmethod
    def add_gvkey_to_crsp(sf):
        """Add gvkey to m(d)sf using ccmxpf_linktable and identify primary stocks.

        There are two tables we can use to link permno with gvkey: ccmxpf_linktable and ccmxpf_linkhist.
        ccmxpf_lnkused is simply a merge table of ccmxpf_lnkhist and ccmxpf_lnkused.

        We identify primary stocks in the following order.

            1. If linkprim = 'P' or 'C', set the security as primary.
            2. If permno and gvkey have 1:1 mapping, set the security as primary.
            3. Among the securities with the same gvkey, set the one with the maximum trading volume as primary.
            4. Among the securities with the same permco and missing gvkey, set the one with the maximum trading volume
               as primary.

        References:
            https://wrds-www.wharton.upenn.edu/pages/support/research-wrds/macros/wrds-macros-cvccmlnksas/

            https://wrds-www.wharton.upenn.edu/pages/support/applications/linking-databases/linking-crsp-and-compustat/

        Args:
            sf: m(d)sf Dataframe with index = 'date'/'permno'.

        Returns:
            m(d)sf with 'gveky' and 'primary' (primary stock indicator) columns added.
        """

        elapsed_time('Adding gvkey to crsp...')

        link = WRDS.read_data('ccmxpf_linktable', 'crsp')
        link = link[~link.lpermno.isna()]
        link = link[link.usedflag == 1]
        link = link[link.linktype.isin(['LC', 'LU', 'LS'])]  # primary links only
        # link = link[link.linktype.isin(['LC', 'LN', 'LU', 'LX', 'LD', 'LS'])]  # primary and secondary links
        link = link[['lpermno', 'gvkey', 'linkdt', 'linkenddt', 'linkprim']].rename(columns={'lpermno': 'permno'})

        # Exceptions
        link.loc[(link.gvkey == '002759') & (link.linkdt == '1985-09-26') & (link.permno == 66843), 'linkprim'] = 'C'
        link.loc[(link.gvkey == '013699') & (link.linkdt == '1988-08-25') & (link.permno == 75189), 'linkprim'] = 'C'
        link.loc[(link.gvkey == '004162') & (link.linkdt == '2002-10-01') & (link.permno == 31318), 'linkprim'] = 'C'

        # To use merge_asof, data should be sorted.
        sf.sort_values('date', inplace=True)
        link.sort_values('linkdt', inplace=True)

        # merge sf with link. linkdt <= date <= linkenddt
        sf['permno'] = sf['permno'].astype('Int64')  # We need this for 'by=permno'.
        link['permno'] = link['permno'].astype('Int64')
        sf = pd.merge_asof(sf, link, left_on='date', right_on='linkdt', by='permno')
        sf.loc[sf.date > sf.linkenddt, ['gvkey', 'linkprim']] = None

        sf['primary'] = False
        sf.loc[sf.linkprim.isin(['P', 'C']), 'primary'] = True
        sf.loc[sf.groupby(['date', 'gvkey']).permno.transform('size') == 1, 'primary'] = True  # Primary if permno and gveky are 1:1

        # Primary identification by volume
        # We first group by gvkey and then by permco since they are not always 1:1.
        sf['dvol'] = sf.prc.abs() * sf.vol  # dollar volume
        sf['max_dvol'] = sf.groupby(['date', 'gvkey'])['dvol'].transform(max)
        sf['nprimary'] = sf.groupby(['date', 'gvkey']).primary.transform(sum)  # num. primary (0 means unidentified)
        sf.loc[(sf.nprimary == 0) & (sf.dvol == sf.max_dvol), 'primary'] = True

        sf['max_dvol'] = sf.groupby(['date', 'permco'])['dvol'].transform(max)
        sf['nprimary'] = sf.groupby(['date', 'permco']).primary.transform(sum)  # num. primary (0 means unidentified)
        sf.loc[sf.gvkey.isna() & (sf.nprimary == 0) & (sf.dvol == sf.max_dvol), 'primary'] = True

        # Check missing or dup
        nprimary = sf.groupby(['date', 'gvkey']).primary.sum()
        if (nprimary != 1).any():
            log('Primary missing or dup.')
            log(nprimary[nprimary != 1])

        sf.drop(columns=['linkdt', 'linkenddt', 'linkprim', 'dvol', 'max_dvol', 'nprimary'], inplace=True)
        sf.sort_values(['permno', 'date'], inplace=True)

        elapsed_time(f'gvkey added to crsp: crsp shape: {sf.shape}, missing gvkey: {sf.gvkey.isna().sum()}')
        log(f'No. unique permnos: {len(sf.permno.unique())}, no. unique gvkeys: {len(sf.gvkey.unique())}')
        return sf

    @staticmethod
    def preprocess_crsp():
        """Create crspm and crspd.

        This method calls ``WRDS.merge_sf_with_seall()`` and ``WRDS.add_gvkey_to_crsp()`` to add delist return,
        gveky, and primary indicator to m(d)sf.
        The result is saved to config.input_dir/crspm(d).pickle.
        """

        crsp = WRDS.read_data('msf')
        # msenames can be missing when a firm is delisted, resulting in missing shrcd/exchcd.
        # => fill with the latest data.
        crsp.sort_values(['permno', 'date'], inplace=True)
        crsp['shrcd'] = crsp.groupby('permno').shrcd.fillna(method='ffill')
        crsp['exchcd'] = crsp.groupby('permno').exchcd.fillna(method='ffill')

        crsp = WRDS.merge_sf_with_seall(crsp, monthly=True)
        crsp = WRDS.add_gvkey_to_crsp(crsp)
        WRDS.save_data(crsp, 'crspm')

        crsp = WRDS.read_data('dsf')
        crsp = WRDS.merge_sf_with_seall(crsp, monthly=False)
        crsp = WRDS.add_gvkey_to_crsp(crsp)
        WRDS.save_data(crsp, 'crspd')

    @staticmethod
    def get_risk_free_rate(sdate=None, edate=None, src='mcti', month_end=False):
        """Get risk-free rate.

        The risk free rate can be obtained either from crsp.mcti or ff.factors_monthly.
        mcti is preferred since the values in factors_monthly have only 4 decimal places.
        Both risk-free rates are in decimal (not percentage values).

        Args:
            sdate: Start date.
            edate: End date.
            src: data source. 'mcti': crsp.mcti, 'ff': ff.factors_monthly.
            month_end: Shift dates to the end of the month.

        Returns:
            Dataframe of risk-free rates with index = 'date' and columns = ['rf'].
        """

        if src == 'mcti':
            rf = WRDS.read_data('mcti')[['caldt', 't30ret']]
            rf = rf.rename(columns={'caldt': 'date', 't30ret': 'rf'}).set_index('date')
        elif src == 'ff':
            rf = WRDS.read_data('factors_monthly')[['dateff', 'rf']]
            rf = rf.rename(columns={'dateff': 'date'}).set_index('date')
        else:
            raise ValueError(f'Undefined risk free data source: {src}')

        if month_end:
            rf.index = to_month_end(rf.index)
        return rf.loc[sdate:edate]

    @staticmethod
    def convert_fund_currency_to_usd(fund, table='funda'):
        """Convert non-USD values of funda(q) to USD values.

        NOTE:
            In Compustat North America, the accounting data can be in either USD and CAD. This is no problem if firm
            characteristics are generated using only Compustat. However, if you mix data from different databases, e.g.,
            if you use market equity of CRSP, which is in USD, Compustat data should be converted to USD.

            Following JKP, we use compustat.exrt_dly to obtain exchange rates. exrt_dly starts from 1982-02-01.

        Args:
            fund: funda(q) DataFrame with index = 'datadate'/'gvkey'.
            table: 'funda' or 'fundq': indicator whether `fund` is funda or fundq.

        Returns:
            Converted `fund` DataFrame.
        """

        curcol = 'curcd' if table == 'funda' else 'curcdq'

        exrt = WRDS.read_data('exrt_dly')
        usd_gbp = exrt.loc[exrt.tocurd == 'USD', ['datadate', 'exratd']]
        exrt = exrt.merge(usd_gbp, on='datadate', how='left')
        exrt['exratd_y'] /= exrt['exratd_x']
        exrt.rename(columns={'tocurd': curcol, 'exratd_y': 'exratd'}, inplace=True)

        index_names = fund.index.names
        fund = fund.reset_index().merge(exrt[['datadate', curcol, 'exratd']], on=['datadate', curcol], how='left')
        fund.loc[fund[curcol] == 'USD', 'exratd'] = 1

        for col in fund:
            if (fund[col].dtype == float) and (col not in ['fyear', 'fyr', 'fyearq', 'fqtr']):
                fund[col] *= fund['exratd']

        return fund.set_index(index_names)

    @staticmethod
    def save_data(data, table, library=None):
        """Save `data` in pickle format.

        We use pickle file format to store data as a pickle file preserves data types and is much faster to read
        compared to a csv file. To convert a pickle file to a csv file, ``WRDS.save_as_csv()`` can be used.
        The `data` is saved in the following location:

            * If `library` = None, config.input_dir/`table`.pickle
            * Otherwise, config.input_dir/`library`/`table`.pickle

        Args:
            data: Data to save (DataFrame).
            table: File name without extension.
            library: Directory.
        """

        fdir = config.input_dir + (library + '/' if library else '')

        if not os.path.isdir(fdir):
            os.makedirs(fdir)

        data.to_pickle(fdir + table + '.pickle')

    @staticmethod
    def read_data(table, library=None, index_col=None, sort_col=None):
        """Read data from config.input_dir/`library`/`table`.pickle.

        `library` argument is redundant as if it is None, all folders under config.input_dir is searched.

        Args:
            table: File name without extension.
            library: Directory.
            index_col: (List of) column(s) to be set as index.
            sort_col: (List of) column(s) to sort data.

        Returns:
            (DataFrame) data read. Index = `index_col`.
        """

        data = None

        if library:
            fpath = config.input_dir + library + '/' + table + '.pickle'
            data = pd.read_pickle(fpath)
        else:  # if library is None, search all subdirectories.
            for dir in os.walk(config.input_dir):
                fpath = dir[0] + '/' + table + '.pickle'
                if os.path.exists(fpath):
                    data = pd.read_pickle(fpath)
        if data is None:
            raise ValueError(f'[{table}] does not exist.')
        if sort_col is not None:
            data.sort_values(sort_col, inplace=True)
        if index_col is not None:
            data.set_index(index_col, inplace=True)

        return data

    @staticmethod
    def save_as_csv(table, library=None, fpath=None):
        """Read `table` and save it to a csv file.

        Args:
            table: File name without extension.
            library: Directory.
            fpath: File path for the csv file. If None, the file is saved to config.input_dir/`library`/`table`.csv
        """

        fpath = fpath or config.input_dir + (library + '/' if library else '') + table + '.csv'
        data = WRDS.read_data(table, library)
        data.to_csv(fpath)

    ################################################################################
    #
    # UNDER CONSTRUCTION
    #
    ################################################################################
    # def download_ibes(self):
    #     log('This function is under construction.')
    #     return
    #
    #     log('downloading ibes...')
    #     sql = """
    #         SELECT
    #            ticker, cusip, fpedats, statpers, anndats_act, numest, anntims_act, medest, actual, stdev
    #         FROM
    #             ibes.statsum_epsus
    #         /*WHERE fpi='6'   1 is for annual forecasts, 6 is for quarterly */
    #         WHERE statpers < anndats_act  /* only keep summarized forecasts prior to earnings annoucement */
    #         AND measure = 'EPS'
    #         AND medest is not null
    #         AND fpedats is not null
    #         AND fpedats >= statpers
    #         ORDER BY cusip, fpedats, statpers
    #         """
    #
    #     data = self.db.raw_sql(sql)
    #     data.fpedats = pd.to_datetime(data.fpedats)
    #     save_data(data, 'ibes')
    #     log(f'download complete.')
    #
    # def add_permno_to_fund(self, fd, dropna=True):
    #     log('This function is under construction.')
    #     return
    #
    #     if 'permno' in fd.columns:
    #         return fd
    #
    #     link = read_data('ccmxpf_linktable', 'crsp')
    #     link = link[~link.lpermno.isna()]
    #     link = link[link.usedflag == 1]
    #     link = link[link.linktype.isin(['LC', 'LN', 'LU', 'LX', 'LD', 'LS'])]
    #     link = link[['lpermno', 'gvkey', 'linkdt', 'linkenddt']].rename(columns={'lpermno': 'permno'})
    #
    #     link.linkdt = pd.to_datetime(link.linkdt)
    #     link.linkenddt = pd.to_datetime(link.linkenddt)
    #
    #     fd = fd.sort_index().reset_index()
    #     link = link.sort_values(['linkdt', 'gvkey'])
    #
    #     fd = pd.merge_asof(fd, link, left_on='datadate', right_on='linkdt', by='gvkey')
    #     # fd = fd[(fd.datadate <= fd.linkenddt) | fd.linkenddt.isna()]
    #     fd.loc[fd.datadate > fd.linkenddt, 'permno'] = None
    #
    #     fd.drop(columns=['linkdt', 'linkenddt'], inplace=True)
    #
    #     fd = fd.sort_values(['gvkey', 'datadate']).set_index(['datadate', 'gvkey'])
    #     if dropna:
    #         log(f'fd size: {len(fd)}')
    #         fd = fd[~fd.permno.isna()]
    #         log(f'fd size with not missing permno: {len(fd)}')
    #     return fd
    #
    # def merge_ibes_crspm(self, ibes, crspm):
    #     """
    #     https://wrds-www.wharton.upenn.edu/pages/support/applications/linking-databases/linking-ibes-and-crsp-data/
    #     * Merging IBES and CRSP datasets using ICLINK table;
    #     proc sql;
    #     create table IBES_CRSP
    #     as select a.ticker, a.STATPERS, a.meanrec, c.permno, c.date, c.ret from ibes.recdsum as a,
    #     home.ICLINK as b,
    #     crsp.msf as c
    #     where a.ticker=b.ticker and b.permno=c.permno and intnx('month',a.STATPERS,0,'E') = intnx('month',c.date,0,'E');
    #     quit;
    #     """
    #     log('This function is under construction.')
    #     return
    #
    #     db = sqlite3.connect(':memory:')
    #
    #     link = read_data('ibcrsphist', 'wrdsapp')
    #
    #     # write the tables
    #     ibes.reset_index()[['cusip', 'fpedats']].to_sql('ibes', db, index=False)
    #     link.to_sql('link', db, index=False)
    #     crspm.reset_index()[['date', 'permno']].to_sql('crspm', db, index=False)
    #     elapsed_time('tables uploaded to db')
    #
    #     # Merge ibes with link
    #     sql = """
    #         CREATE TABLE temp as
    #         SELECT l.lpermno as permno, f.*
    #         FROM
    #             ibes as f, link as l
    #         /*ON f.cusip = l.lcusip
    #         AND f.fpedats >= sdate
    #         AND f.fpedats <= edate
    #         WHERE l.lpermno is not null*/
    #         """
    #
    #     db.execute(sql)
    #     elapsed_time('funda and link merged')
    #
    #     sql = """
    #         SELECT c.*, f.*
    #         FROM
    #             crspm as c
    #         LEFT JOIN
    #             temp as f
    #         ON c.permno = f.permno
    #         AND c.date >= f.datadate + 7
    #         AND c.date < f.datadate + 20
    #         """
    #
    #     table = pd.read_sql_query(sql, db)
    #     return table


if __name__ == '__main__':
    os.chdir('../')

    wrds = WRDS('fehouse')
    # wrds.download_all()
    # wrds.preprocess_crsp()
    wrds.download_table('ff', 'factors_monthly', date_cols=['date', 'dateff'])





"""This module defines utility functions."""

from collections.abc import Iterable
from pyanomaly.globals import *


def is_iterable(x):
    """Check if `x` is iterable.

    A string, a iterable object, is considered not iterable in this function.

    Returns:
         True if `x` is iterable.
    """

    return (not isinstance(x, str)) and isinstance(x, Iterable)


def to_list(x):
    """Convert `x` to a list.

    Args:
        x: a list, tuple, or scalar.

    Returns:
        `x` converted to list.
    """

    if is_iterable(x):
        return list(x)
    else:
        return [x]


def unique_list(x, exclude_nan=True):
    """Get a list of unique elements of `x`.

    If `x` contains a list, its elements are considered elements of `x`, not the list itself.

    >>> x = [1, 2, 2, None]
    >>> unique_list(x)
    [1, 2]
    >>> x = [1, 2, [1, 3], None]
    >>> unique_list(x)
    [1, 2, 3]

    Args:
        x: A list.
        exclude_nan: If True, exclude None elements.

    Returns:
        List of unique elements of `x`.
    """

    y = []
    for i in x:
        if is_iterable(i):
            for j in i:
                y.append(j)
        else:
            y.append(i)

    y = list(set(y))
    if exclude_nan and None in y:
        y.remove(None)
    return y


def drop_columns(data, cols):
    """Delete `cols` columns from `data`."""

    for col in cols:
        del data[col]


def keep_columns(data, cols):
    """Delete columns of `data` except `cols`.

    Equivalent to `data = data[cols]`, but much more memory efficient.
    Not sure why but the following two methods seem to momentarily consume a lot of memory:

    >>> data = data[cols]
    >>> data.drop(columns=data.columns.difference(cols), inplace=True)

    Use this function when handling a large dataset.

    Args:
        data: DataFrame.
        cols: List of columns to keep.
    """

    del_cols = data.columns.difference(cols)
    drop_columns(data, del_cols)


ZERO = 1.e-7


def is_zero(x):
    """Check if `x` is zero.

    A value is considered 0 if it is in the range (-1.e-7, 1.e-7).

    Args:
        x: A scalar or array.
    """

    return (x < ZERO) & (x > -ZERO)


"""This module defines functions to generate firm characteristics, factors, and characteristic portfolios
replicating JKP's SAS code.
"""

from pyanomaly.globals import *
# config.REPLICATE_JKP = True

from pyanomaly.analytics import *
from pyanomaly.fileio import write_to_file, read_from_file
from pyanomaly.characteristics import CRSPM, CRSPD, FUNDQ, FUNDA, Merge

################################################################################
#
# Firm characteristics generation
#
################################################################################

def generate_firm_characterisitcs(fname=None):
    """Firm characteristics generation (Replication of JKP's SAS code).

    This function generate the firm characteristics that appear in JKP (2021).
    We replicate JKP's SAS code as closely as possible.
    The output is saved to 'config.output_dir/fname.pickle`.

    It is assumed that

        - the raw data has been downloaded from WRDS. If not, call ``WRDS.download_all()``;
        - factor portfolios has been created. If not, call ``make_factor_portfolios()``.

    Args:
        fname: Output file name. If None, fname = 'merge'.
    """

    set_log_path('./log/generate_firmcharacterisitcs.log')
    elapsed_time()

    # Generate characteristics in 'jkp' column in mapping.xlsx.
    alias = 'jkp'
    # Start date.
    sdate = '1950-01-01'

    ###################################
    # CRSPM
    ###################################
    drawline()
    log('PROCESSING CRSPM')
    drawline()

    # Load raw data.
    crspm = CRSPM(alias=alias)
    crspm.load_data(sdate)

    # Preprocess data.
    crspm.filter_data()
    crspm.populate(freq=MONTHLY, method=None)
    crspm.update_variables()
    crspm.merge_with_factors()

    # Generate characteristics.
    crspm_chars = crspm.get_available_chars()
    crspm.create_chars(crspm_chars)

    # Save the output.
    crspm.postprocess()
    # crspm.save()

    ###################################
    # CRSPD
    ###################################
    drawline()
    log('PROCESSING CRSPD')
    drawline()

    # Load raw data.
    crspd = CRSPD(alias=alias)
    crspd.load_data(sdate)

    # Preprocess data.
    crspd.filter_data()
    crspd.update_variables()
    crspd.merge_with_factors()

    # Generate characteristics.
    crspd_chars = crspd.get_available_chars()
    crspd.create_chars(crspd_chars)

    # Save the output.
    crspd.postprocess()
    # crspd.save()

    ###################################
    # FUNDQ
    ###################################
    drawline()
    log('PROCESSING FUNDQ')
    drawline()

    # Load raw data.
    fundq = FUNDQ(alias=alias)
    fundq.load_data(sdate)

    # Preprocess data.
    fundq.remove_duplicates()
    fundq.convert_currency()
    fundq.convert_to_monthly()
    fundq.create_qitems_from_yitems()
    fundq.update_variables()

    # Generate characteristics.
    fundq_chars = fundq.get_available_chars()
    fundq.create_chars(fundq_chars)

    # Save the output.
    fundq.postprocess()
    # fundq.save()

    ###################################
    # FUNDA
    ###################################
    drawline()
    log('PROCESSING FUNDA')
    drawline()

    # Load raw data.
    funda = FUNDA(alias=alias)
    funda.load_data(sdate)

    # Preprocess data.
    funda.convert_currency()
    funda.convert_to_monthly()
    funda.merge_with_fundq(fundq)
    funda.update_variables()
    funda.add_crsp_me(crspm)

    # Generate characteristics.
    funda_chars = funda.get_available_chars()
    funda.create_chars(funda_chars)

    # Save the output.
    funda.postprocess()
    # funda.save()

    ###################################
    # Merge
    ###################################
    drawline()
    log('PROCESSING MERGE')
    drawline()

    # Merge all data together.
    merge = Merge()
    merge.preprocess(crspm, crspd, funda, fundq)

    # Generate characteristics.
    merge_chars = merge.get_available_chars()
    merge.create_chars(merge_chars)

    # Save the output.
    merge.postprocess()
    merge.save(fname)


################################################################################
#
# Factor portfolio generation
#
################################################################################

def prepare_data_for_factors(monthly=True, sdate=None):
    """Create characteristics needed to make factors.

    Args:
        monthly: If True (False), generate data monthly (daily).
        sdate: Start date ('yyyy-mm-dd').
    """

    crspm = CRSPM()
    crspm.load_data(sdate)

    crspm.filter_data()
    crspm.update_variables()

    fundq = FUNDQ()
    fundq.load_data(sdate)

    fundq.remove_duplicates()
    fundq.convert_currency()
    fundq.convert_to_monthly()
    fundq.create_qitems_from_yitems()
    fundq.update_variables()

    fundq.create_chars(['niq_be'])

    funda = FUNDA()
    funda.load_data(sdate)

    funda.convert_currency()
    funda.convert_to_monthly()
    funda.merge_with_fundq(fundq)
    funda.update_variables()
    funda.add_crsp_me(crspm)

    funda.create_chars(['at_gr1', 'be_me'])

    if not monthly:
        crspd = CRSPD()
        crspd.load_data(sdate)

        crspd.filter_data()
        crspd.update_variables()

        keep_columns(crspd.data, ['gvkey', 'rf', 'exchcd', 'primary', 'me', 'me_company', 'exret', 'ym'])
        data = crspd.data.reset_index()

        data = data.merge(funda.data[['at_gr1', 'be_me']], left_on=['ym', 'gvkey'], right_on=['date', 'gvkey'], how='left')
        data = data.merge(fundq.data[['niq_be']], left_on=['ym', 'gvkey'], right_on=['date', 'gvkey'], how='left')
    else:
        keep_columns(crspm.data, ['gvkey', 'rf', 'exchcd', 'primary', 'me', 'me_company', 'exret'])
        data = crspm.data.reset_index()

        data = data.merge(funda.data[['at_gr1', 'be_me']], on=['date', 'gvkey'], how='left')
        data = data.merge(fundq.data[['niq_be']], on=['date', 'gvkey'], how='left')

    return data.set_index(['date', 'permno']).sort_index(level=[1, 0])


def make_factor_portfolio(data, char, ret_col, size_class, weight_col=None):
    """Make a factor portfolio.

    Args:
        data: Dataframe with index = date/id. Its columns should include `char`, `ret_col`, `size_class`, and
            `weight_col` (optional).
        char: Characteristic column to make a factor portfolio from.
        ret_col: Return column.
        size_class: Size class column.
        weight_col: Weight column. If None, stocks are equally weighted.

    Returns:
        Dataframe of (size x `char`) factor portfolios. Index = 'date',
        columns: ['bh', 'bl', 'bm', 'sh', 'sl', 'sm', 'hml', 'smb'].
    """

    char_split = [0.3, 0.7, 1.0]
    char_by = char + '_by'  # Column to determine quantile breakpoints: same as char but keeps only NYSE stocks.
    char_class = char + '_class'  # Class (group) column.

    # Classify stocks on `char`.
    data[char_by] = np.where(data.exchcd == 1, data[char], np.nan)  # Keep only NYSE stocks.
    data[char_class] = classify(data[char], char_split, ascending=False, by_array=data[char_by])

    # 2-D sort. factors index = date/char_class/size_class, columns = [ret].
    factors = two_dim_sort(data, char_class, size_class, ret_col, weight_col=weight_col, output_dim=2)

    # Name portfolios.
    relabel_class(factors,  ['h', 'm', 'l', 'hml'], axis=0)
    relabel_class(factors, ['s', 'b', 'smb'], axis=1)
    factors = factors.unstack()
    factors.columns = factors.columns.map(''.join)

    factors['hml'] = (factors['shml'] + factors['bhml']) / 2
    factors['smb'] = (factors['smbh'] + factors['smbm'] + factors['smbl']) / 3
    drop_columns(factors, ['shml', 'bhml', 'smbh', 'smbm', 'smbl', 'smbhml'])

    elapsed_time(f'factor portfolio created: {char}')

    return factors


def make_factor_portfolios(monthly=True, sdate=None):
    """Make factor portfolios.

    FF 3 and HXZ 4 factors are generated. The result is saved to `config.factors_monthly(daily)_fname`.

    Args:
        monthly: If True (False), generate data monthly (daily).
        sdate: Start date ('yyyy-mm-dd').

    Returns:
        Factor portfolio dataframe with index = 'date' and columns = ['mktrf', 'rf', 'smb_ff', 'hml', 'inv', 'roe',
        'smb_hxz'].
    """
    elapsed_time('make_factor_portfolios')

    data = prepare_data_for_factors(monthly, sdate)
    # data = read_jkp_data()

    ret_col = 'futret'  # t+1 excess return
    size_col = 'me' # We use security level me following JKP but think we should use me_company.

    # One-period ahead return.
    data[ret_col] = make_future_return(data.exret)
    # The futret at t is the return from t to t+1. Shift one period so that the return is from t-1 to t and
    # other variables are as of t.
    data = data.groupby(level=-1).shift(1)  # Shift data forward.

    ########################################
    # Filtering
    ########################################
    data = data.dropna(subset=[ret_col, size_col])  # non-missing return and me
    data = data[data.primary == True]  # primary securities only

    ########################################
    # Size split
    ########################################
    size_split = [0.5, 1.0]
    size_class = 'size_class'

    data['nyse_me'] = np.where(data.exchcd == 1, data[size_col], np.nan)  # Keep only NYSE stocks.
    data[size_class] = classify(data[size_col], size_split, ascending=True, by_array=data['nyse_me'])

    ########################################
    # Factor portfolios
    ########################################
    # mMrket pf.
    factors = weighted_mean(data, ret_col, size_col, 'date')
    factors.rename(columns={ret_col: 'mktrf'}, inplace=True)
    factors['rf'] = data.groupby('date').rf.first()

    # FF factors.
    hml = make_factor_portfolio(data, 'be_me', ret_col, size_class, size_col)
    factors[['smb_ff', 'hml']] = hml[['smb', 'hml']]

    # HXZ factors.
    inv = make_factor_portfolio(data, 'at_gr1', ret_col, size_class, size_col)
    roe = make_factor_portfolio(data, 'niq_be', ret_col, size_class, size_col)
    factors['inv'] = -inv['hml']
    factors['roe'] = roe['hml']
    factors['smb_hxz'] = (inv['smb'] + roe['smb']) / 2

    # Save results.
    if monthly:
        write_to_file(factors, config.factors_monthly_fname)
    else:
        write_to_file(factors, config.factors_daily_fname)

    return factors


################################################################################
#
# Characteristic portfolios
#
################################################################################
def make_char_portfolios(data, char_list, weighting):
    """Make characteristic portfolios.

    Args:
        data: DataFrame of firm characteristics.
        char_list: List of characteristics to generate.
        weighting: 'ew' (Equal-weight), 'vw' (Value-weight), or 'vw_cap' (Value-weight capped at 0.8 NYSE-size quantile).

    Returns:
        Characteristic portfolio DataFrame with index = date/class and columns = [char, 'ret', 'signal', 'n_firms'].
        The class values are one of 'h', 'm', 'l', 'hml'.
    """

    elapsed_time(f'make_char_portfolios. weighting: {weighting}')

    ret_col = 'futret'  # t+1 excess return
    size_col = 'me'  # We use security level me following JKP but think we should use me_company.

    # One-period ahead return.
    data[ret_col] = make_future_return(data.exret)
    # The futret at t is the return from t to t+1. Shift one period so that the return is from t-1 to t and
    # other variables are as of t.
    data = data.groupby(level=-1).shift(1)  # Shift data forward.

    ########################################
    # Filtering
    ########################################
    data = data.dropna(subset=[ret_col, size_col])  # non-missing t+1 return and me
    data = data[data.primary == True]  # primary securities only

    ########################################
    # Size split
    ########################################
    data['nyse_me'] = np.where(data.exchcd == 1, data[size_col], np.nan)  # NYSE me

    if weighting == 'vw':
        weight_col = size_col
    elif weighting == 'vw_cap':
        data['me_cap'] = winsorize(data[size_col], (None, 0.2), by_array=data['nyse_me'])
        weight_col = 'me_cap'
    else: # weighting == 'ew':
        weight_col = None

    elapsed_time('size_class generated.')

    ########################################
    # Characteristic portfolios
    ########################################
    # Add identifier for the bottom 20%.
    size_split = [0.2, 1.0]
    size_class = 'size_class'
    data[size_class] = classify(data[size_col], size_split, ascending=True, by_array=data['nyse_me'])

    split = 3
    char_by = 'char_by'
    group_col = 'group'
    labels = ['h', 'm', 'l', 'hml']

    char_portfolios = []
    for char in char_list:
        data[char_by] = np.where(data[size_class] > 0, data[char], np.nan)  # NYSE-characteristic

        # Classify data on char.
        data[group_col] = classify(data[char], split, ascending=False, by_array=data[char_by])

        # Make quantile portfolios.
        portfolios = one_dim_sort(data, group_col, ret_col, weight_col=weight_col)
        portfolios[['signal', 'n_firms']] = one_dim_sort(data, group_col, char, function=['mean', 'count'])

        relabel_class(portfolios, labels)
        portfolios = portfolios.rename(columns={ret_col: 'ret'})
        portfolios['characteristic'] = char
        char_portfolios.append(portfolios.reset_index(group_col))

        elapsed_time(f'Characteristic portfolio created: {char}')

    return pd.concat(char_portfolios)


def make_char_portfolios_all():
    alias = 'jkp'

    data = read_from_file('merge_full')

    # Choose characteristics.
    char_map = pd.read_excel(config.mapping_file_path)
    char_list = list(char_map['jkp'].dropna())
    char_list = [char for char in char_list if char in data]  # Only those in data.

    # Equal-weight
    char_portfolios = make_char_portfolios(data, char_list, 'ew')
    write_to_file(char_portfolios, 'char_portfolios_ew')

    # Value-weight
    char_portfolios = make_char_portfolios(data, char_list, 'vw')
    write_to_file(char_portfolios, 'char_portfolios_vw')

    # Capped value-weight
    char_portfolios = make_char_portfolios(data, char_list, 'vw_cap')
    write_to_file(char_portfolios, 'char_portfolios_vw_cap')


################################################################################
#
# Aux. function for testing. (IGNORE)
#
# The fuctions below may not work on your side as they read some files I created
# for testing purposes.
################################################################################

def read_jkp_data():
    # Read JKP data.

    data = pd.read_pickle('jkp/kelly_2000.pickle')
    # data = pd.read_pickle('jkp/world_data.pickle')
    # data.date = to_month_end(data.date)
    data = data.set_index(['date', 'permno'])
    data['exret'] = data.ret_exc
    # data['futret'] = data.groupby('permno').ret_exc.shift(-1)
    data = data.rename(columns={'primary_sec': 'primary', 'crsp_exchcd': 'exchcd'})
    data['primary'] = data['primary'].astype(bool)
    data = data[(data.common == 1) & (data.ret_lag_dif == 1) & (data.obs_main == 1)]

    return data


def covert_jkp_factor_portfolios():
    # Convert factor files generated by JKP's SAS code to pickle files.

    fdata = pd.read_csv('AP_FACTORS_MONTHLY.csv')
    fdata = fdata[fdata.excntry == 'USA']
    fdata['date'] = pd.to_datetime(fdata.eom.astype('str'))
    fdata.set_index('date', inplace=True)
    fdata.to_pickle('output/factors_monthly_jkp.pickle')

    fdata = pd.read_csv('AP_FACTORS_DAILY.csv')
    fdata = fdata[fdata.excntry == 'USA']
    fdata['date'] = pd.to_datetime(fdata.date.astype('str'))
    fdata.set_index('date', inplace=True)
    fdata.to_pickle('output/factors_daily_jkp.pickle')


def test_factor_portfolios(monthly=True):
    # Test factor portfolio generation.

    factors = make_factor_portfolios(monthly=monthly)

    if monthly:
        factors_jkp = read_from_file('factors_monthly_jkp')
    else:
        factors_jkp = read_from_file('factors_daily_jkp')

    compare_data(factors, factors_jkp, tolerance=0.1)
    return factors


def test_char_portfolio():
    # Test characteristic portfolio generation.

    char_list = ['ret_1_0', 'ret_12_1']

    data = read_jkp_data()  # use data generated by jkp

    # JKP result
    hml_jkp = pd.read_csv('jkp/hml_pfs.csv')
    hml_jkp['date'] = pd.to_datetime(hml_jkp.eom.astype(str))
    hml_jkp = hml_jkp.set_index('date')
    hml_jkp = hml_jkp[hml_jkp.characteristic.isin(char_list)]

    hml = None
    for weighting in ['vw', 'ew', 'vw_cap']:
        char_pfs = make_char_portfolios(data, char_list, weighting)
        if hml is None:
            hml = char_pfs[char_pfs['group'] == 'hml']
            hml = hml.rename(columns={'ret': 'ret_' + weighting})
        else:
            hml['ret_' + weighting] = char_pfs.loc[char_pfs['group'] == 'hml', 'ret']

    for char in char_list:
        print('\n', char)
        compare_data(hml[hml.characteristic==char], hml_jkp[hml_jkp.characteristic == char], tolerance=0.1)


if __name__ == '__main__':
    os.chdir('../')

    # make_factor_portfolios(monthly=True)
    # make_factor_portfolios(monthly=False)
    # generate_firm_characterisitcs()
    make_char_portfolios_all()

    # test_factor_portfolios(monthly=False)
    # test_char_portfolio()




"""This module defines functions for file IO."""

from .globals import *

def write_to_file(data, fname, fdir=None):
    """Write `data` to a pickle file.

    Args:
        data (Dataframe): Data to save.
        fname: File name without extension.
        fdir: Directory. If None, `config.output_dir`.

    """
    fdir = fdir or config.output_dir

    if not os.path.isdir(fdir):
        os.mkdir(fdir)

    data.to_pickle(fdir + ('' if fdir[-1] == '/' else '/') + fname + '.pickle')


def read_from_file(fname, fdir=None):
    """Read data from a pickle file.

    Args:
        fname: File name without extension.
        fdir: Directory. If None, `config.output_dir`.

    Returns:
        DataFrame read from `fdir/fname.pickle`.
    """
    fdir = fdir or config.output_dir
    fpath = fdir + ('' if fdir[-1] == '/' else '/') + fname + '.pickle'
    return pd.read_pickle(fpath)






import numpy as np
import pandas as pd
from concurrent.futures import ProcessPoolExecutor, as_completed
from pyanomaly.globals import *

def process(i, module, fcn, data, *args, **kwargs):
    log(f'process {i} started.')
    kwargs['mp'] = False
    s = fcn.split('.')
    exec(f'from {module} import {s[0]}')  # import the relevant class or function
    return i, eval(fcn)(data, *args, **kwargs)

def multiprocess(func):
    max_prc = 10

    def wrapper(data, *args, **kwargs):
        if 'mp' in kwargs and not kwargs['mp']:
            kwargs.pop('mp')
            return func(data, *args, **kwargs)

        firms_list = np.array_split(data.index.get_level_values(-1).unique(), max_prc * 4)

        date_col = data.index.names[0]
        data = data.reset_index(level=0)
        char_list = [0] * len(firms_list)
        futures = []
        with ProcessPoolExecutor(max_workers=max_prc) as executor:
            for i, firms in enumerate(firms_list):
                data_ = data.loc[firms].set_index(date_col, append=True).swaplevel()
                futures.append(executor.submit(process, i, func.__module__, func.__qualname__, data_, *args, **kwargs))

            for f in as_completed(futures):
                i, char = f.result()
                char_list[i] = char
                log(f'Multiprocessing complete: {i}/{len(firms_list)}.')

        if isinstance(char_list[0], tuple):  # multiple return values
            n_var = len(char_list[0])
            retval = [[] for _ in range(n_var)]
            for i in range(n_var):
                for char in char_list:
                    retval[i].append(char[i])
                retval[i] = np.concatenate(retval[i])
        elif isinstance(char_list[0], np.ndarray):  # single return value (np.array)
            retval = np.concatenate(char_list)
        elif isinstance(char_list[0], pd.DataFrame):  # single return value (pd.DataFrame)
            retval = pd.concat(char_list)

        elapsed_time('Multiprocessing: Results gathered.')
        return retval

    return wrapper





import numpy as np
from numba import njit


@njit(error_model='numpy')
def std(a, ddof=1):
    n = a.shape[0]
    s = a.std()
    return s if ddof == 0 else s * np.sqrt(n / (n - 1))


@njit(error_model='numpy')
def skew(a, ddof=1):
    n = a.shape[0]
    m = a.mean()
    s = std(a, ddof)
    a = (a - m) / s

    return (a ** 3).mean() if ddof == 0 else (a ** 3).sum() * n / (n-1) / (n-2)

@njit
def add_constant(a):
    return np.hstack((np.ones((a.shape[0], 1)), a))


@njit
def isnan1(data):
    """Equivalent to np.isnan(axis=1)"""
    isnan = np.full(data.shape[0], False)
    for j in range(data.shape[1]):
        isnan |= np.isnan(data[:, j])
    return isnan


